import math
from dataclasses import dataclass
from typing import Dict, List, Set, Tuple

import pytest

from python_ta.contracts import check_all_contracts


def is_valid_name(name):
    return name.isalpha()


class Person:
    """
    Represent a person.

    Representation Invariants:
    - self.age > 0
    - len(self.name) > 0
    - is_valid_name(self.name)
    """

    age: int
    name: str
    fav_foods: List[str]

    def __init__(self, name, age, fav_food):
        self.name = name
        self.age = age
        self.fav_foods = fav_food

    def change_name(self, name: str) -> str:
        self.name = name
        return name

    def change_age(self, age: int) -> int:
        """
        Precondition: age < 150
        """
        self.age = age
        return age

    def decrease_and_increase_age(self, age: int) -> int:
        self.age = -10
        self.age = age
        return age

    def add_fav_food(self, food):
        self.fav_foods.append(food)


def change_age(person, new_age):
    person.age = new_age


class Pizza:
    """
    Representation Invariants:
    - len(self.ingredients) > 0
    - 0 \
        < self.radius \
            <= 10
    """

    def __init__(self, radius, ingredients):
        self.radius = radius
        self.ingredients = ingredients

    @classmethod
    def margherita(cls, radius):
        return cls(radius, ["mozzarella", "tomatoes"])

    def area(self):
        return self.circle_area(self.radius)

    @staticmethod
    def circle_area(r):
        """
        Precondition: r > 0
        """
        return r ** 2 * math.pi


@dataclass
class SetWrapper:
    """A wrapper around a set.

    Representation Invariants:
        - all(x in self.set for x in {1, 2, 3})
    """

    set: Set[int]


@pytest.fixture
def person():
    return Person("David", 31, ["Sushi"])


def test_change_age_invalid_over(person) -> None:
    """
    Change the age to larger than 100. Expect an exception.
    """
    with pytest.raises(AssertionError) as excinfo:
        person.change_age(200)
    msg = str(excinfo.value)
    assert "age < 150" in msg


def test_change_name_valid(person) -> None:
    """
    Change the name using a valid name.
    """
    person.name = "Ignas"
    assert person.name == "Ignas"


def test_change_name_invalid_nonalpha(person) -> None:
    """
    Change name to contain non-alphabet characters. Expect an exception.
    """
    with pytest.raises(AssertionError) as excinfo:
        person.name = "$$"
    msg = str(excinfo.value)
    assert "is_valid_name(self.name)" in msg


def test_change_age_invalid_in_method(person) -> None:
    """
    Call a method that changes age to something invalid but back to something valid.
    Expects normal behavior.
    """
    age = person.decrease_and_increase_age(10)
    assert age == 10


def test_same_method_names(person) -> None:
    """
    Call a method with the same name as an instance method and ensure reprsentation invariants are checked.
    Expects normal behavior.
    """

    with pytest.raises(AssertionError) as excinfo:
        change_age(person, -10)
    msg = str(excinfo.value)
    assert "self.age > 0" in msg


def test_wrong_food_type(person) -> None:
    """Change fav food to an int. Expect an exception."""

    with pytest.raises(AssertionError):
        person.fav_foods = 5


def test_wrong_food_type_instance_method(person) -> None:
    """Violates type annotation within an instance method. Expect exception."""

    with pytest.raises(AssertionError):
        person.add_fav_food(5)


def test_create_margherita_invalid() -> None:
    """
    Create circle area with invalid r. Also tests multiline conditions.
    Expect an exception.
    """
    with pytest.raises(AssertionError) as excinfo:
        Pizza.margherita(0)
    msg = str(excinfo.value)
    assert (
        "0 \
        < self.radius \
            <= 10"
        in msg
    )


def test_circle_area_valid() -> None:
    """
    Calculate circle area with valid r.
    """
    a = Pizza.circle_area(10)
    assert a == (10 ** 2 * math.pi)


def test_circle_area_invalid() -> None:
    """
    Calculate circle area with invalid r. Expect an exception.
    """
    with pytest.raises(AssertionError) as excinfo:
        Pizza.circle_area(0)
    msg = str(excinfo.value)
    assert "r > 0" in msg


def test_set_wrapper_valid() -> None:
    """
    Test the SetWrapper representation invariant on a valid instance.
    """
    my_set = SetWrapper(set={1, 2, 3})
    assert my_set.set == {1, 2, 3}


def test_set_wrapper_invalid() -> None:
    """
    Test the SetWrapper representation invariant on a valid instance.
    """
    with pytest.raises(AssertionError) as excinfo:
        SetWrapper(set={1, 2, -3})
    msg = str(excinfo.value)
    assert "all(x in self.set for x in {1, 2, 3})" in msg


class NoInit:
    """A class with no initializer.

    Representation Invariants:
        - abs(1) < 0  # This is always False
    """

    def method(self) -> int:
        """Method to test that representation invariant is checked on method calls."""
        return 1


def test_no_init_setattr() -> None:
    """
    Check that a built-in function (abs) can be called successfully
    from a representation invariant of a class with no __init__ method, when setting an attribute.
    """
    with pytest.raises(AssertionError) as excinfo:
        n = NoInit()
        n.attr = 1

    msg = str(excinfo.value)
    assert "abs(1) < 0" in msg


def test_no_init_method() -> None:
    """
    Check that a built-in function (abs) can be called successfully
    from a representation invariant of a class with no __init__ method, when calling a method.
    """
    with pytest.raises(AssertionError) as excinfo:
        n = NoInit()
        n.method()

    msg = str(excinfo.value)
    assert "abs(1) < 0" in msg


class NoInit2:
    """A class with no initializer.

    Representation Invariants:
        - is_valid_name('123')  # This is always False
    """

    def method(self) -> int:
        """Method to test that representation invariant is checked on method calls."""
        return 1


def test_no_init_setattr2() -> None:
    """
    Check that a user-defined function (is_valid_name) can be called successfully
    from a representation invariant of a class with no __init__ method, when setting an attribute.
    """
    with pytest.raises(AssertionError) as excinfo:
        n = NoInit2()
        n.attr = 1

    msg = str(excinfo.value)
    assert "is_valid_name('123')" in msg


def test_no_init_method2() -> None:
    """
    Check that a user-defined function (is_valid_name) can be called successfully
    from a representation invariant of a class with no __init__ method, when calling a method.
    """
    with pytest.raises(AssertionError) as excinfo:
        n = NoInit2()
        n.method()

    msg = str(excinfo.value)
    assert "is_valid_name('123')" in msg


def setup_module() -> None:
    """Pytest hook for setting up the module"""
    check_all_contracts(__name__, decorate_main=False)


class OverrideLen:
    """A class that overrides its len method."""

    x: int

    def __init__(self) -> None:
        self.x = 1

    def __len__(self) -> int:
        return self.x


def test_override_len_simple() -> None:
    """Test that we can instantiate OverrideLen, which has a custom __len__ implementation."""
    OverrideLen()


class ThemedWidget:
    """
    Representation Invariants:
    - self.theme.lower() in {'dark', 'light'}
    - self.primary_color.isalpha()
    - self.secondary_color.isalpha()
    """

    size: int
    theme: str
    primary_color: str
    secondary_color: str

    def __init__(self, theme: str, color_palette: Tuple[str, str], options: Dict = None) -> None:
        if options:
            self.setup_options(options)
        else:
            self.size = 5
        self.apply_theme(theme)
        self.apply_color_palette(color_palette)

    def setup_options(self, options: Dict) -> None:
        if "size" in options:
            self.setup_size(options["size"])

    def setup_size(self, size: int) -> None:
        self.size = size

    def apply_theme(self, theme: str) -> None:
        self.theme = theme

    def apply_color_palette(self, color_palette: Tuple[str, str]) -> None:
        self.primary_color, self.secondary_color = color_palette


def test_no_premature_check_from_helper_in_init() -> None:
    """Test that representation invariants and type annotations of a class still being
    initialized are not checked when a helper called directly from the init returns
    """
    dark_widget = ThemedWidget("dark", ("black", "mahogany"))
    assert dark_widget.theme == "dark"
    assert dark_widget.primary_color == "black"
    assert dark_widget.secondary_color == "mahogany"


def test_no_premature_check_from_deep_helper_in_init() -> None:
    """Test that representation invariants and type annotations of a class still being
    initialized are not checked when a helper of any depth from the init returns
    """
    dark_widget = ThemedWidget("dark", ("black", "mahogany"), options={"size": 10})
    assert dark_widget.theme == "dark"
    assert dark_widget.primary_color == "black"
    assert dark_widget.secondary_color == "mahogany"


if __name__ == "__main__":
    pytest.main(["test_class_contracts.py"])



from typing import *
from typing import Callable, ForwardRef

from pytest import skip

from python_ta.typecheck.base import (
    TypeConstraints,
    TypeFail,
    TypeInfo,
    TypeResult,
    create_Callable,
)

skip_msg = "Skipped"
tc = TypeConstraints()


# Helper functions
def unify_helper(arg1: type, arg2: type, exp_result: Union[type, TypeFail]):
    unify_result = TypeInfo(arg1) >> (lambda t1: TypeInfo(arg2) >> (lambda t2: tc.unify(t1, t2)))
    if isinstance(exp_result, TypeFail):
        assert isinstance(unify_result, TypeFail)
    else:
        assert tc.resolve(unify_result.getValue()).getValue() == tc.resolve(exp_result).getValue()


def setup_typevar(t: type):
    tv = tc.fresh_tvar(None)
    tc.unify(tv, t)
    return tv


def resolve_helper(t: type, exp_type: type):
    assert tc.resolve(t).getValue() == exp_type


# Unify primitives
def test_same_prim():
    unify_helper(bool, bool, bool)
    unify_helper(int, int, int)
    unify_helper(str, str, str)


def test_diff_prim():
    unify_helper(bool, str, TypeFail(f"Incompatible Types {bool} and {str}"))
    unify_helper(int, str, TypeFail(f"Incompatible Types {int} and {str}"))
    unify_helper(bool, int, TypeFail(f"Incompatible Types {bool} and {int}"))
    unify_helper(float, int, TypeFail(f"Incompatible Types {float} and {int}"))
    unify_helper(float, str, TypeFail(f"Incompatible Types {float} and {str}"))


# Unify TypeVars
def test_same_typevars():
    tc.reset()
    tv1 = setup_typevar(str)
    tv2 = setup_typevar(str)

    resolve_helper(tv1, str)
    resolve_helper(tv2, str)

    unify_helper(tv1, tv2, tv1)


def test_same_typevars_flipped():
    tc.reset()
    tv1 = setup_typevar(str)
    tv2 = setup_typevar(str)

    resolve_helper(tv1, str)
    resolve_helper(tv2, str)
    unify_helper(tv1, tv2, tv2)


def test_diff_typevars():
    tc.reset()
    tv_str = setup_typevar(str)
    tv_int = setup_typevar(int)

    resolve_helper(tv_str, str)
    resolve_helper(tv_int, int)
    unify_helper(tv_int, tv_str, TypeFail(f"Incompatible Types {str} and {int}"))


def test_one_typevar():
    tc.reset()
    tv = setup_typevar(str)

    resolve_helper(tv, str)
    unify_helper(tv, str, str)
    unify_helper(str, tv, str)
    unify_helper(tv, int, TypeFail(f"Incompatible Types {str} and {int}"))
    unify_helper(int, tv, TypeFail(f"Incompatible Types {int} and {str}"))


def test_two_typevar():
    tc.reset()
    tv1 = setup_typevar(bool)
    tv2 = tc.fresh_tvar(None)
    unify_helper(tv1, tv2, bool)


# Unify ForwardRefs
def test_same_forward_ref():
    fr1 = ForwardRef("a")
    fr2 = ForwardRef("a")
    unify_helper(fr1, fr2, fr1)
    unify_helper(fr1, fr2, fr2)


def test_diff_forward_ref():
    skip("The existing error msg does not apply to this situation")
    fr1 = ForwardRef("a")
    fr2 = ForwardRef("b")
    unify_helper(fr1, fr2, TypeFail("Attempted to unify forwardref  with non-ref"))


def test_one_forward_ref():
    fr = ForwardRef("a")
    unify_helper(fr, str, TypeFail("Attempted to unify forwardref  with non-ref"))


# Unify Tuples
def test_same_tuple():
    unify_helper(Tuple[int, int], Tuple[int, int], Tuple[int, int])
    unify_helper(Tuple[str, str], Tuple[str, str], Tuple[str, str])


def test_diff_tuple():
    unify_helper(
        Tuple[int, int],
        Tuple[str, str],
        TypeFail(f"Incompatible Types {Tuple[int, int]} and {Tuple[str, str]}"),
    )


def test_nested_tuples():
    unify_helper(
        Tuple[str, Tuple[str, bool]], Tuple[str, Tuple[str, bool]], Tuple[str, Tuple[str, bool]]
    )


def test_typevars_tuple():
    tv1 = tc.fresh_tvar(None)
    tv2 = tc.fresh_tvar(None)
    unify_helper(Tuple[tv1, tv2], Tuple[str, bool], Tuple[str, bool])
    resolve_helper(tv1, str)
    resolve_helper(tv2, bool)


def test_typevars_nested_tuples():
    tv1 = tc.fresh_tvar(None)
    tv2 = Tuple[tv1, bool]
    unify_helper(tv2, Tuple[Tuple[str, bool], bool], Tuple[Tuple[str, bool], bool])
    resolve_helper(tv1, Tuple[str, bool])
    resolve_helper(tv2, Tuple[Tuple[str, bool], bool])


def test_diff_nested_tuples():
    unify_helper(
        Tuple[str, Tuple[str, str]],
        Tuple[str, Tuple[bool, str]],
        TypeFail(
            f"Incompatible Types {Tuple[str, Tuple[str, str]]} and {Tuple[str, Tuple[bool, str]]}"
        ),
    )


# Unify list
def test_same_list():
    unify_helper(List[str], List[str], List[str])
    unify_helper(List[int], List[int], List[int])


def test_diff_list():
    unify_helper(List[str], List[int], TypeFail(f"Incompatible Types {List[str]} and {List[int]}"))


# Unify callables
def test_same_callable():
    tc.reset()
    c1 = Callable[[bool], bool]
    c2 = Callable[[bool], bool]

    unify_helper(c1, c2, c1)
    unify_helper(c1, c2, c2)
    unify_helper(c2, c1, c1)
    unify_helper(c2, c1, c2)


def test_diff_callable():
    c1 = Callable[[bool], bool]
    c2 = Callable[[str], str]

    unify_helper(c1, c2, TypeFail(f"Incompatible Types {c1} and {c2}"))


# Polymorphic types
def test_simple_polymorphic_call():
    tc.reset()
    tv1 = tc.fresh_tvar()
    tv2 = tc.fresh_tvar()
    fn1 = create_Callable([tv1, tv2], bool, {tv1, tv2})
    fn2 = create_Callable([int, int], bool)

    unify_helper(fn1, fn2, Callable[[int, int], bool])


def test_higher_order_polymorphic_call():
    tc.reset()
    tv1 = tc.fresh_tvar()
    tv2 = tc.fresh_tvar()

    fn0 = create_Callable([tv1, int], int, {tv1})
    fn1 = create_Callable([int, int], int)

    fn2 = create_Callable([fn0, int], bool)
    fn3 = create_Callable([fn1, int], bool)
    fn4 = create_Callable([tv2, int], bool, {tv2})

    unify_helper(fn2, fn3, Callable[[Callable[[int, int], int], int], bool])
    unify_helper(fn2, fn4, Callable[[Callable[[int, int], int], int], bool])

    resolve_helper(tv1, int)
    resolve_helper(tv2, Callable[[int, int], int])


# Union types
def test_simple_union():
    tc.reset()
    unify_helper(int, Union[int, str], int)


def test_tvar_union():
    tc.reset()
    unify_helper(tc.fresh_tvar(), Union[int, str], Union[int, str])
    unify_helper(Union[int, str], tc.fresh_tvar(), Union[int, str])


def test_two_unions():
    tc.reset()
    unify_helper(Union[int, str, None], Union[bool, int], int)


def test_optional():
    tc.reset()
    tv1 = tc.fresh_tvar()

    unify_helper(int, Optional[int], int)
    unify_helper(type(None), Optional[int], type(None))
    unify_helper(tv1, Optional[int], Optional[int])



"""Tests for setendings.py.

Checks that `end_lineno` and `end_col_offset` node properties are set.
"""
import os.path as path
import sys
import unittest

import astroid

from python_ta.transforms.setendings import *

# Gives absolute path to root directory
PATH = path.normpath(path.join(path.abspath(__file__), "..", "..", "examples", "ending_locations"))


class TestEndingLocations(unittest.TestCase):
    """The method, ending_transformer.visit(module) walks the given astroid
    *tree* and transform each encountered node. Only the nodes which have
    transforms registered will actually be replaced or changed.

    We store the correct values as a tuple:
    (fromlineno, end_lineno, col_offset, end_col_offset)
    """

    def get_file_as_module(self, file_location):
        """Given a filepath (file_location), parse with astroid, and return
        the module.
        """
        file = path.join(PATH, file_location)
        with open(file) as f:
            content = f.read()
        return self.get_string_as_module(content)

    def get_string_as_module(self, string):
        """Parse the string with astroid, and return the module.

        Also initialize the ending transformer here.
        """
        source_lines = string.split("\n")
        # Instantiate a visitor, and register the transform functions to it.
        self.ending_transformer = init_register_ending_setters(source_lines)

        return astroid.parse(string)

    def set_and_check(self, module, node_class, expected):
        """Example is either in a file, or provided as a string literal."""
        self.ending_transformer.visit(module)  # Apply all transforms.
        props = [
            (node.fromlineno, node.end_lineno, node.col_offset, node.end_col_offset)
            for node in module.nodes_of_class(node_class)
        ]
        self.assertEqual(expected, props)

    def test_arguments(self):
        expected = [(1, 2, 8, 30), (5, 5, 14, 14), (8, 8, 12, 12), (9, 9, 14, 18), (11, 11, 6, 48)]
        module = self.get_file_as_module("arguments.py")
        self.set_and_check(module, nodes.Arguments, expected)

    def test_assert(self):
        expected = [(1, 1, 0, 43), (2, 2, 0, 11)]
        module = self.get_file_as_module("assert.py")
        self.set_and_check(module, nodes.Assert, expected)

    def test_assign(self):
        expected = [(1, 1, 0, 5), (2, 2, 0, 9), (3, 3, 0, 11), (4, 4, 0, 8), (5, 5, 0, 6)]
        module = self.get_file_as_module("assign.py")
        self.set_and_check(module, nodes.Assign, expected)

    def test_assignattr(self):
        """
        Given 'self.name = 10', we want to highlight 'self.name' rather than
        just 'self'.
        """
        expected = [(3, 3, 8, 17), (4, 4, 8, 19)]
        module = self.get_file_as_module("assign_attr.py")
        self.set_and_check(module, nodes.AssignAttr, expected)

    def test_assignname(self):
        expected = [(1, 1, 0, 1), (2, 2, 0, 1), (2, 2, 4, 5), (3, 3, 0, 1), (3, 3, 3, 4)]
        module = self.get_file_as_module("assign_name.py")
        self.set_and_check(module, nodes.AssignName, expected)

    def test_asyncfor(self):
        """Note: col_offset property always set after the 'async' keyword."""
        expected = [(3, 7, 4, 16)]
        module = self.get_file_as_module("async_for.py")
        self.set_and_check(module, nodes.AsyncFor, expected)

    def test_asyncfunctiondef(self):
        expected = [(4, 5, 0, 27)]
        module = self.get_file_as_module("async_function_def.py")
        self.set_and_check(module, nodes.AsyncFunctionDef, expected)

    def test_asyncwith(self):
        expected = [(2, 3, 4, 12)]
        module = self.get_file_as_module("async_with.py")
        self.set_and_check(module, nodes.AsyncWith, expected)

    def test_attribute(self):
        """Note: Setting the attribute node by its last child doesn't include
        the attribute in determining the end_col_offset.
        """
        expected = [(1, 1, 0, 12), (2, 2, 0, 14)]
        module = self.get_file_as_module("attribute.py")
        self.set_and_check(module, nodes.Attribute, expected)

    def test_augassign(self):
        expected = [(1, 1, 0, 6)]
        module = self.get_file_as_module("aug_assign.py")
        self.set_and_check(module, nodes.AugAssign, expected)

    def test_await(self):
        expected = [(5, 5, 4, 27)]
        module = self.get_file_as_module("await.py")
        self.set_and_check(module, nodes.Await, expected)

    def test_binop(self):
        expected = [
            (1, 1, 0, 9),
            (1, 1, 0, 5),
            (2, 2, 0, 13),
            (3, 3, 0, 17),
            (4, 4, 0, 19),
            (4, 4, 1, 8),
            (4, 4, 11, 18),
        ]
        module = self.get_file_as_module("bin_op.py")
        self.set_and_check(module, nodes.BinOp, expected)

    def test_boolop(self):
        expected = [(1, 1, 4, 13), (2, 2, 0, 32), (2, 2, 0, 23)]
        module = self.get_file_as_module("bool_op.py")
        self.set_and_check(module, nodes.BoolOp, expected)

    def test_break(self):
        expected = [(2, 2, 4, 9)]
        module = self.get_file_as_module("break.py")
        self.set_and_check(module, nodes.Break, expected)

    def test_call(self):
        expected = [(1, 2, 0, 9)]
        module = self.get_file_as_module("call.py")
        self.set_and_check(module, nodes.Call, expected)

    def test_classdef(self):
        expected = [(1, 3, 0, 8)]
        module = self.get_file_as_module("class_def.py")
        self.set_and_check(module, nodes.ClassDef, expected)

    def test_compare(self):
        expected = [(1, 1, 0, 10), (2, 2, 0, 5), (3, 3, 0, 12), (4, 4, 0, 26)]
        module = self.get_file_as_module("compare.py")
        self.set_and_check(module, nodes.Compare, expected)

    def test_comprehension(self):
        """
        Could be in a SetComp, ListComp, or GeneratorExp. In each respective
        case, the subsequent char could be either a brace, bracket, or paren.
        Aside: col_offset should start from beginning of the 'for'.
        """
        expected = [(1, 1, 7, 20), (2, 2, 7, 16), (2, 2, 21, 36), (3, 3, 9, 18), (3, 3, 23, 40)]
        module = self.get_file_as_module("comprehension.py")
        self.set_and_check(module, nodes.Comprehension, expected)

    def test_const(self):
        expected = [
            (1, 1, 0, 6),
            (2, 2, 4, 6),
            (3, 3, 0, 3),
            (4, 4, 0, 8),
            (5, 7, 0, 1),
            (8, 8, 6, 11),
            (8, 8, 13, 25),
            (9, 9, 0, 3),
        ]
        module = self.get_file_as_module("const.py")
        self.set_and_check(module, nodes.Const, expected)

    def test_continue(self):
        expected = [(2, 2, 4, 12)]
        module = self.get_file_as_module("continue.py")
        self.set_and_check(module, nodes.Continue, expected)

    def test_decorators(self):
        """
        Include the right parens (note: only if decorator takes args)
        """
        expected = [(1, 2, 0, 27), (6, 6, 0, 9)]
        module = self.get_file_as_module("decorators.py")
        self.set_and_check(module, nodes.Decorators, expected)

    def test_delattr(self):
        """Include the 'del' keyword in the col_offset property.
        Include the attribute name in the end_col_offset property.
        """
        expected = [(4, 4, 8, 21), (5, 5, 8, 23)]
        module = self.get_file_as_module("del_attr.py")
        self.set_and_check(module, nodes.DelAttr, expected)

    def test_delete(self):
        """Include the 'del' keyword in the col_offset property."""
        expected = [(1, 1, 0, 5), (2, 2, 0, 22)]
        module = self.get_file_as_module("delete.py")
        self.set_and_check(module, nodes.Delete, expected)

    def test_delname(self):
        """Include the 'del' keyword in the col_offset property."""
        expected = [(1, 1, 0, 5)]
        module = self.get_file_as_module("del_name.py")
        self.set_and_check(module, nodes.DelName, expected)

    def test_dict(self):
        expected = [(1, 1, 6, 32), (2, 5, 4, 1), (6, 9, 4, 6)]
        module = self.get_file_as_module("dict.py")
        self.set_and_check(module, nodes.Dict, expected)

    def test_dictcomp(self):
        expected = [(1, 1, 0, 29), (2, 2, 0, 37), (3, 7, 0, 1)]
        module = self.get_file_as_module("dict_comp.py")
        self.set_and_check(module, nodes.DictComp, expected)

    # def test_dictunpack(self):
    #     """NODE EXAMPLE DOES NOT EXIST
    #     """
    #     expected = []
    #     module = self.get_file_as_module('DictUnpack.py')
    #     self.set_and_check(module, nodes.DictUnpack, expected)

    # def test_emptynode(self):
    #     """NODE EXAMPLE DOES NOT EXIST
    #     """
    #     expected = []
    #     module = self.get_file_as_module('EmptyNode.py')
    #     self.set_and_check(module, nodes.EmptyNode, expected)

    def test_excepthandler(self):
        expected = [(3, 4, 0, 8)]
        module = self.get_file_as_module("except_handler.py")
        self.set_and_check(module, nodes.ExceptHandler, expected)

    # def test_exec(self):
    #     """NODE EXAMPLE DOES NOT EXIST
    #     """
    #     expected = []
    #     module = self.get_file_as_module('Exec.py')
    #     self.set_and_check(module, nodes.Exec, expected)

    def test_expr(self):
        expected = [(1, 1, 0, 12), (2, 2, 0, 13), (3, 3, 0, 11), (4, 4, 0, 17), (5, 5, 0, 31)]
        module = self.get_file_as_module("expr.py")
        self.set_and_check(module, nodes.Expr, expected)

    def test_for(self):
        expected = [(1, 4, 0, 8)]
        module = self.get_file_as_module("for.py")
        self.set_and_check(module, nodes.For, expected)

    def test_functiondef(self):
        """Note: this node includes the decorator."""
        expected = [(1, 7, 0, 28)]
        module = self.get_file_as_module("function_def.py")
        self.set_and_check(module, nodes.FunctionDef, expected)

    def test_generatorexp(self):
        expected = [(1, 1, 0, 37), (2, 2, 0, 43)]
        module = self.get_file_as_module("generator_exp.py")
        self.set_and_check(module, nodes.GeneratorExp, expected)

    def test_global(self):
        expected = [(2, 2, 4, 15)]
        module = self.get_file_as_module("global.py")
        self.set_and_check(module, nodes.Global, expected)

    def test_if(self):
        """Note: each elif is represented as a separate If node."""
        expected = [(1, 8, 0, 9), (3, 8, 0, 9), (5, 8, 0, 9)]
        module = self.get_file_as_module("if.py")
        self.set_and_check(module, nodes.If, expected)

    def test_ifexp(self):
        expected = [(1, 1, 4, 20)]
        module = self.get_file_as_module("if_exp.py")
        self.set_and_check(module, nodes.IfExp, expected)

    def test_import(self):
        expected = [(1, 1, 0, 21), (2, 2, 0, 30), (3, 3, 0, 28)]
        module = self.get_file_as_module("import.py")
        self.set_and_check(module, nodes.Import, expected)

    def test_importfrom(self):
        expected = [(1, 1, 0, 46), (2, 2, 0, 65), (3, 3, 0, 31), (4, 4, 0, 32)]
        module = self.get_file_as_module("import_from.py")
        self.set_and_check(module, nodes.ImportFrom, expected)

    def test_keyword(self):
        """Include the name of the keyword, contained in 'node.arg' attribute."""
        expected = [(1, 1, 4, 12), (2, 2, 5, 15)]
        module = self.get_file_as_module("keyword.py")
        self.set_and_check(module, nodes.Keyword, expected)

    def test_lambda(self):
        expected = [(1, 1, 6, 15), (2, 2, 7, 25)]
        module = self.get_file_as_module("lambda.py")
        self.set_and_check(module, nodes.Lambda, expected)

    def test_list(self):
        expected = [(1, 1, 0, 2), (2, 2, 0, 9), (3, 3, 0, 6), (4, 9, 0, 1)]
        module = self.get_file_as_module("list.py")
        self.set_and_check(module, nodes.List, expected)

    def test_listcomp(self):
        expected = [(1, 1, 0, 24), (2, 2, 0, 49)]
        module = self.get_file_as_module("list_comp.py")
        self.set_and_check(module, nodes.ListComp, expected)

    def test_module(self):
        expected = [(0, 3, 0, 11)]
        module = self.get_file_as_module("module.py")
        self.set_and_check(module, nodes.Module, expected)

    def test_name(self):
        expected = [(1, 1, 0, 6)]
        module = self.get_file_as_module("name.py")
        self.set_and_check(module, nodes.Name, expected)

    def test_nonlocal(self):
        expected = [(4, 4, 8, 21)]
        module = self.get_file_as_module("nonlocal.py")
        self.set_and_check(module, nodes.Nonlocal, expected)

    def test_pass(self):
        expected = [(3, 3, 8, 12)]
        module = self.get_file_as_module("pass.py")
        self.set_and_check(module, nodes.Pass, expected)

    # def test_print(self):
    #     """NODE EXAMPLE DOES NOT EXIST
    #     """
    #     expected = []
    #     module = self.get_file_as_module('Print.py')
    #     self.set_and_check(module, nodes.Print, expected)

    def test_raise(self):
        expected = [(3, 3, 8, 24), (5, 5, 8, 36)]
        module = self.get_file_as_module("raise.py")
        self.set_and_check(module, nodes.Raise, expected)

    # def test_repr(self):
    #     """NODE EXAMPLE DOES NOT EXIST
    #     """
    #     expected = []
    #     module = self.get_file_as_module('Repr.py')
    #     self.set_and_check(module, nodes.Repr, expected)

    def test_return(self):
        expected = [(3, 3, 8, 14), (5, 5, 8, 42)]
        module = self.get_file_as_module("return.py")
        self.set_and_check(module, nodes.Return, expected)

    def test_set(self):
        expected = [(1, 1, 0, 3), (2, 2, 0, 6), (3, 3, 0, 12)]
        module = self.get_file_as_module("set.py")
        self.set_and_check(module, nodes.Set, expected)

    def test_setcomp(self):
        expected = [(1, 1, 0, 25), (2, 2, 0, 63)]
        module = self.get_file_as_module("set_comp.py")
        self.set_and_check(module, nodes.SetComp, expected)

    def test_slice(self):
        """Note: includes square brackets."""
        expected = [
            (1, 1, 1, 4),
            (2, 2, 2, 9),
            (3, 3, 1, 5),
            (4, 4, 2, 14),
            (5, 5, 1, 9),
            (6, 6, 7, 31),
            (7, 8, 1, 3),
            (9, 9, 1, 5),
            (9, 9, 5, 8),
            (10, 10, 1, 4),
            (10, 10, 4, 8),
            (11, 11, 1, 4),
            (11, 11, 4, 7),
            (12, 12, 1, 5),
            (13, 13, 1, 4),
            (13, 13, 9, 12),
            (14, 14, 5, 12),
            (15, 15, 1, 4),
            (15, 15, 4, 7),
            (15, 15, 8, 11),
        ]
        module = self.get_file_as_module("slice.py")
        self.set_and_check(module, nodes.Slice, expected)

    def test_starred(self):
        expected = [(1, 1, 0, 2), (4, 4, 6, 8)]
        module = self.get_file_as_module("starred.py")
        self.set_and_check(module, nodes.Starred, expected)

    def test_subscript(self):
        expected = [
            (1, 1, 0, 4),
            (2, 2, 0, 8),
            (3, 3, 0, 4),
            (4, 4, 0, 9),
            (5, 5, 0, 5),
            (6, 6, 0, 14),
            (7, 7, 0, 9),
            (8, 8, 4, 31),
            (9, 10, 0, 3),
            (11, 11, 0, 8),
            (11, 11, 0, 5),
            (12, 12, 0, 8),
            (12, 12, 0, 4),
            (13, 13, 0, 7),
            (13, 13, 0, 4),
            (14, 14, 0, 5),
            (15, 15, 0, 12),
            (15, 15, 0, 4),
            (16, 16, 4, 12),
            (22, 22, 17, 26),
            (22, 22, 31, 40),
        ]
        module = self.get_file_as_module("subscript.py")
        self.set_and_check(module, nodes.Subscript, expected)

    def test_tryexcept(self):
        expected = [(1, 4, 0, 8)]
        module = self.get_file_as_module("try_except.py")
        self.set_and_check(module, nodes.TryExcept, expected)

    def test_tryfinally(self):
        expected = [(1, 6, 0, 8)]
        module = self.get_file_as_module("try_finally.py")
        self.set_and_check(module, nodes.TryFinally, expected)

    def test_tuple(self):
        expected = [
            (1, 1, 0, 6),
            (2, 2, 0, 11),
            (3, 3, 0, 5),
            (4, 4, 0, 7),
            (5, 5, 0, 12),
            (6, 8, 0, 8),
            (9, 9, 0, 4),
            (10, 10, 0, 7),
            (11, 13, 0, 17),
            (14, 14, 0, 6),
            (15, 15, 7, 13),
            (16, 16, 4, 10),
            (17, 17, 0, 10),
            (17, 17, 0, 4),
            (17, 17, 6, 10),
            (18, 18, 0, 6),
            (20, 20, 0, 6),
            (21, 21, 0, 2),
        ]
        module = self.get_file_as_module("tuple.py")
        self.set_and_check(module, nodes.Tuple, expected)

    def test_unaryop(self):
        expected = [(1, 1, 0, 8), (2, 2, 0, 2), (3, 3, 0, 2), (4, 4, 0, 3)]
        module = self.get_file_as_module("unary_op.py")
        self.set_and_check(module, nodes.UnaryOp, expected)

    def test_while(self):
        expected = [(1, 5, 0, 13)]
        module = self.get_file_as_module("while.py")
        self.set_and_check(module, nodes.While, expected)

    def test_with(self):
        expected = [(1, 3, 0, 19)]
        module = self.get_file_as_module("with.py")
        self.set_and_check(module, nodes.With, expected)

    def test_yield(self):
        expected = [(2, 2, 4, 11)]
        module = self.get_file_as_module("yield.py")
        self.set_and_check(module, nodes.Yield, expected)

    def test_yieldfrom(self):
        expected = [(2, 2, 4, 23)]
        module = self.get_file_as_module("yield_from.py")
        self.set_and_check(module, nodes.YieldFrom, expected)


if __name__ == "__main__":
    unittest.main()  # run tests



import sys
from typing import Callable, ForwardRef, Generic, Type, _GenericAlias

import hypothesis.strategies as hs
import pytest
from astroid import nodes
from hypothesis import HealthCheck, assume, given, settings

from python_ta.transforms.type_inference_visitor import TypeFail
from python_ta.typecheck.base import _gorg

from .. import custom_hypothesis_support as cs
from ..custom_hypothesis_support import lookup_type, types_in_callable

settings.load_profile("pyta")


def _parse_to_function(function_name, args_list, return_statement):
    """Helper to parse given data into function definition."""
    return f'def {function_name}({", ".join(args_list)}):' f"    return {return_statement}"


def _parse_to_function_no_return(function_name, args_list, function_body):
    """Helper to parse given data into function definition."""
    return f'def {function_name}({", ".join(args_list)}):\n' f"    {function_body}"


@pytest.mark.skip("This test is flaky")
@given(cs.valid_identifier(), hs.lists(cs.valid_identifier(), min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_inference_args_simple_return(function_name, arguments):
    """Test whether visitor was able to infer type of argument given function called on it in function body."""
    assume(function_name not in arguments)
    for argument in arguments:
        program = _parse_to_function_no_return(
            function_name, arguments, ("return " + argument + " + " + repr("bob"))
        )
        module, inferer = cs._parse_text(program)
        # get the functionDef node - there is only one in this test case.
        function_def_node = next(module.nodes_of_class(nodes.FunctionDef))
        expected_arg_type_vars = [
            function_def_node.type_environment.lookup_in_env(argument) for argument in arguments
        ]
        expected_arg_types = [
            inferer.type_constraints.resolve(type_var).getValue()
            for type_var in expected_arg_type_vars
        ]
        function_type_var = module.type_environment.lookup_in_env(function_name)
        function_type = inferer.type_constraints.resolve(function_type_var).getValue()
        actual_arg_types, actual_return_type = types_in_callable(inferer, function_type)
        assert expected_arg_types == actual_arg_types


@given(cs.valid_identifier(), hs.lists(cs.valid_identifier(), min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_function_def_args_simple_return(function_name, arguments):
    """Test whether visitor was able to infer type of function given function called on its arguments."""
    assume(function_name not in arguments)
    for argument in arguments:
        program = _parse_to_function_no_return(
            function_name, arguments, ("return " + argument + " + " + repr("bob"))
        )
        module, inferer = cs._parse_text(program)
        function_def_node = next(module.nodes_of_class(nodes.FunctionDef))
        expected_arg_type_vars = [
            function_def_node.type_environment.lookup_in_env(argument) for argument in arguments
        ]
        expected_arg_types = [
            inferer.type_constraints.resolve(type_var).getValue()
            for type_var in expected_arg_type_vars
        ]
        function_type_var = module.type_environment.lookup_in_env(function_name)
        function_type = inferer.type_constraints.resolve(function_type_var).getValue()
        actual_arg_types, actual_return_type = types_in_callable(inferer, function_type)
        return_type_var = function_def_node.type_environment.lookup_in_env(argument)
        expected_return_type = inferer.type_constraints.resolve(return_type_var).getValue()
        assert (
            Callable[actual_arg_types, actual_return_type]
            == Callable[expected_arg_types, expected_return_type]
        )


@given(cs.functiondef_node(annotated=True, returns=True))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_functiondef_annotated_simple_return(functiondef_node):
    """Test whether type annotations are set properly for a FunctionDef node representing a function definition
    with type annotations."""
    arg_names = [arg.name for arg in functiondef_node.args.args]
    assume(functiondef_node.name not in arg_names)
    for arg in functiondef_node.args.args:
        assume(arg_names.count(arg.name) == 1)
    module, inferer = cs._parse_text(functiondef_node)
    functiondef_node = next(module.nodes_of_class(nodes.FunctionDef))
    # arguments and annotations are not changing, so test this once.
    for i in range(len(functiondef_node.args.annotations)):
        arg_name = functiondef_node.args.args[i].name
        expected_type = inferer.type_constraints.resolve(
            functiondef_node.type_environment.lookup_in_env(arg_name)
        ).getValue()
        # need to do by name because annotations must be name nodes.
        if isinstance(expected_type, _GenericAlias):
            assert _gorg(expected_type).__name__ == functiondef_node.args.annotations[i].name
        elif sys.version_info >= (3, 9) and hasattr(expected_type, "__origin__"):
            assert expected_type.__origin__.__name__ == functiondef_node.args.annotations[i].name
        else:
            assert expected_type.__name__ == functiondef_node.args.annotations[i].name
    # test return type
    return_node = functiondef_node.body[0].value
    expected_rtype = inferer.type_constraints.resolve(
        functiondef_node.type_environment.lookup_in_env(return_node.name)
    ).getValue()
    if isinstance(expected_rtype, _GenericAlias):
        assert _gorg(expected_rtype).__name__ == functiondef_node.returns.name
    elif sys.version_info >= (3, 9) and hasattr(expected_rtype, "__origin__"):
        assert expected_rtype.__origin__.__name__ == functiondef_node.args.annotations[i].name
    else:
        assert expected_rtype.__name__ == functiondef_node.returns.name


def test_functiondef_method():
    program = """
        class A:

            def method(self, x):
                return x + 1
        """
    module, inferer = cs._parse_text(program)
    for func_def in module.nodes_of_class(nodes.FunctionDef):
        assert lookup_type(inferer, func_def, func_def.argnames()[0]) == ForwardRef("A")


def test_functiondef_classmethod():
    program = """
        class A:

            @classmethod
            def method(cls, x):
                return x + 1
        """
    module, inferer = cs._parse_text(program)
    for func_def in module.nodes_of_class(nodes.FunctionDef):
        assert lookup_type(inferer, func_def, func_def.argnames()[0]) == Type[ForwardRef("A")]


def test_functiondef_staticmethod():
    program = """
        class A:

            @staticmethod
            def method(x):
                return x + 1
        """
    module, inferer = cs._parse_text(program)
    for func_def in module.nodes_of_class(nodes.FunctionDef):
        assert lookup_type(inferer, func_def, func_def.argnames()[0]) == int


def test_nested_annotated_function_conflicting_body():
    """User tries to define an annotated function which has conflicting types within its body."""
    program = f"def random_func(int1: int) -> None:\n" f'    int1 + "bob"\n'

    module, inferer = cs._parse_text(program)
    binop_node = next(module.nodes_of_class(nodes.BinOp))
    assert isinstance(binop_node.inf_type, TypeFail)


def test_annotated_functiondef_conflicting_return_type():
    """User defines an annotated function with type errors in it's body;
    a discrepancy in annotated return type versus return type in it's body.
    """
    program = (
        f"def return_str(num1: int, str1: str) -> int:\n"
        f"    output = num1 + str1\n"
        f'    return "bob"\n'
        f"\n"
    )
    module, inferer = cs._parse_text(program)
    functiondef_node = next(module.nodes_of_class(nodes.FunctionDef))
    assert isinstance(functiondef_node.inf_type, TypeFail)


def test_function_return():
    program = """
    def foo(x):
        return x

    foo(1)
    """
    module, inferer = cs._parse_text(program)
    call_node = next(module.nodes_of_class(nodes.Call))
    func_type = call_node.func.inf_type.getValue()
    t1, t2 = func_type.__args__
    assert t1 == t2


def test_function_return_2():
    program = """
    def foo(x, y):
        return x

    foo(1,2)
    """
    module, inferer = cs._parse_text(program)
    call_node = next(module.nodes_of_class(nodes.Call))
    func_type = call_node.func.inf_type.getValue()
    t1, t2, t3 = func_type.__args__
    assert t1 == t3


def test_function_return_3():
    program = """
    def foo(x, y):
        return y

    foo(1,2)
    """
    module, inferer = cs._parse_text(program)
    call_node = next(module.nodes_of_class(nodes.Call))
    func_type = call_node.func.inf_type.getValue()
    t1, t2, t3 = func_type.__args__
    assert t2 == t3



from typing import Dict, Set

import os
import subprocess
import re
import pytest
import json
import itertools


_EXAMPLES_PATH = "examples/pylint/"
_EXAMPLE_PREFIX_REGEX = r"[CEFRW]\d{4}"


# The following tests appear to always fail (further investigation needed).
IGNORED_TESTS = [
    "E1131_unsupported_binary_operation.py",
    "E0118_used_prior_global_declaration.py",
    "W0125_using_constant_test.py",
    "W0631_undefined_loop_variable.py",
    "W1503_redundant_unittest_assert.py",
    "E1140_unhashable_dict_key.py",
    "R0401_cyclic_import.py",  # R0401 required an additional unit test but should be kept here.
]


def get_file_paths():
    """Gets all the files from the examples folder for testing. This will
    return all the full file paths to the file, meaning they will have the
    _EXAMPLES_PATH prefix followed by the file name for each element.
    A list of all the file paths will be returned."""
    test_files = []
    for _, _, files in os.walk(_EXAMPLES_PATH, topdown=True):
        for filename in files:
            if filename not in IGNORED_TESTS and filename.endswith(".py"):
                test_files.append(_EXAMPLES_PATH + filename)
    return test_files


@pytest.fixture(scope="session", autouse=True)
def symbols_by_file() -> Dict[str, Set[str]]:
    """Run pylint on all the example files and return the map of file name to the
    set of Pylint messages it raises."""
    output = subprocess.run(
        [
            "pylint",
            "--reports=n",
            "--rcfile=python_ta/.pylintrc",
            "--output-format=json",
            *get_file_paths(),
        ],
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
    )

    jsons_output = output.stdout.decode("UTF-8")
    pylint_list_output = json.loads(jsons_output)

    file_to_symbol = {}
    for path, group in itertools.groupby(pylint_list_output, key=lambda d: d["path"]):
        symbols = {message["symbol"] for message in group}
        file = os.path.basename(path)

        file_to_symbol[file] = symbols

    return file_to_symbol


@pytest.mark.parametrize("test_file", get_file_paths())
def test_examples_files(test_file: str, symbols_by_file: Dict[str, Set[str]]) -> None:
    """Creates all the new unit tests dynamically from the testing directory."""
    base_name = os.path.basename(test_file)
    if not re.match(_EXAMPLE_PREFIX_REGEX, base_name[:5]):
        return
    if not base_name.lower().endswith(".py"):
        assert False
    checker_name = base_name[6:-3].replace("_", "-")  # Take off prefix and file extension.

    test_file_name = os.path.basename(test_file)
    file_symbols = symbols_by_file[test_file_name]

    found_pylint_message = checker_name in file_symbols
    assert found_pylint_message, f"Failed {test_file}. File does not add expected message."


def test_cyclic_import() -> None:
    """Test that examples/pylint/R0401_cyclic_import adds R0401 cyclic-import.

    Reason for creating a separate test:
    This test is separate as pylint adds the R0401 message to the final module within
    the batch of given modules, not to the R0401_cyclic_import or cyclic_import_helper.
    It is unintuitive to force R0401_cyclic_import to be the final batched module so that
    the parametrized test suite passes, so cyclic-import is ignored in the paramterized suite
    and this additional test is created on the side.
    """

    cyclic_import_helper = "examples/pylint/cyclic_import_helper.py"
    cyclic_import_file = "examples/pylint/R0401_cyclic_import.py"

    output = subprocess.run(
        [
            "pylint",
            "--reports=n",
            "--rcfile=python_ta/.pylintrc",
            "--output-format=json",
            *[cyclic_import_helper, cyclic_import_file],
        ],
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
    )

    jsons_output = output.stdout.decode("UTF-8")
    pylint_list_output = json.loads(jsons_output)

    found_cyclic_import = any(
        message["symbol"] == "cyclic-import" for message in pylint_list_output
    )
    assert found_cyclic_import, f"Failed {cyclic_import_file}. File does not add expected message."



from typing import List

from astroid import nodes
from hypothesis import HealthCheck, given, settings
from pytest import skip

from python_ta.transforms.type_inference_visitor import NoType
from python_ta.typecheck.base import TypeFail, TypeFailFunction

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


@given(cs.subscript_node(cs.simple_homogeneous_list_node(min_size=1)))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_inference_list_subscript(node):
    """Test whether visitor properly set the type constraint of Subscript node representing list-index access."""
    module, _ = cs._parse_text(node)
    for subscript_node in module.nodes_of_class(nodes.Subscript):
        list_node = subscript_node.value
        assert subscript_node.inf_type.getValue() == list_node.elts[0].inf_type.getValue()


@given(cs.subscript_node(cs.simple_homogeneous_list_node(min_size=1), cs.slice_node()))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_subscript_homogeneous_list_slice(node):
    """Test visitor of Subscript node representing slicing of homogeneous list."""
    module, _ = cs._parse_text(node)
    for subscript_node in module.nodes_of_class(nodes.Subscript):
        list_node = subscript_node.value
        assert subscript_node.inf_type.getValue() == List[list_node.elts[0].inf_type.getValue()]


@given(cs.subscript_node(cs.simple_homogeneous_list_node(min_size=1), cs.slice_node()))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_subscript_load_ctx(node):
    """Test visitor of Subscript node when loaded in an (if) expression."""
    load_node = nodes.If()
    load_node.postinit(nodes.Const(True), [node], [])
    module, _ = cs._parse_text(load_node)
    for subscript_node in module.nodes_of_class(nodes.Subscript):
        list_node = subscript_node.value
        assert subscript_node.inf_type.getValue() == List[list_node.elts[0].inf_type.getValue()]


def test_homogenous_list_store_ctx():
    """Test visitor of Subscript node within a homogenous list assignment."""
    program = """
        [1,2,3][0] = 2
        """
    module, _ = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.Assign):
        assert assign_node.inf_type == NoType()
    for subscript_node in module.nodes_of_class(nodes.Subscript):
        assert subscript_node.inf_type == NoType()


def test_homogenous_list_invalid_store_ctx():
    """Test visitor of Subscript node within an invalid homogenous list assignment."""
    program = """
        [1,2,3][0] = 'a'
        """
    module, _ = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.Assign):
        assert isinstance(assign_node.inf_type, TypeFail)
    for subscript_node in module.nodes_of_class(nodes.Subscript):
        assert subscript_node.inf_type == NoType()


@given(cs.subscript_node(cs.list_node(min_size=1), cs.slice_node()))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_subscript_del_ctx(node):
    """Test visitor of Subscript node within a del statement."""
    del_node = nodes.Delete()
    del_node.postinit([node])
    module, _ = cs._parse_text(del_node)
    for subscript_node in module.nodes_of_class(nodes.Subscript):
        assert subscript_node.inf_type.getValue() == type(None)


@given(cs.simple_homogeneous_dict_node(min_size=1))
def test_inference_dict_subscript(node):
    """Note that this test only takes in a dictionary because the subscript index
    must be the same type as the dictionary's keys in order to type check.
    """
    for key, _ in node.items:
        new_node = nodes.Subscript()
        new_node.postinit(node, key)
        module, _ = cs._parse_text(new_node)
        for subscript_node in module.nodes_of_class(nodes.Subscript):
            dict_node = subscript_node.value
            assert (
                subscript_node.inf_type.getValue()
                == list(dict_node.items)[0][1].inf_type.getValue()
            )


@given(cs.simple_homogeneous_list_node(min_size=1))
def test_inference_invalid_slice(node):
    sub_node = nodes.Subscript()
    slice = nodes.Slice()
    slice.postinit(nodes.Const(0), nodes.Const("a"))
    sub_node.postinit(node, slice)
    module, _ = cs._parse_text(sub_node)
    for subscript_node in module.nodes_of_class(nodes.Subscript):
        assert isinstance(subscript_node.inf_type, TypeFail)


def test_inference_ext_slice():
    skip("Lookup for class methods must be implemeneted before this test can pass")
    program = """
        class Foo:
            def __getitem__(self, tup):
                return tup[0]

        foo = Foo()
        foo[1, 'a']
        """
    module, _ = cs._parse_text(program)
    subscript_node = list(module.nodes_of_class(nodes.Subscript))[1]
    assert subscript_node.inf_type.getValue() == int


def test_subscript_slice():
    program = """
        x = List[:]
        """
    module, _ = cs._parse_text(program)
    assign_node = next(module.nodes_of_class(nodes.Assign))
    assert isinstance(assign_node.inf_type, TypeFail)


# TODO: this test needs to be converted, but will also fail
# @given(cs.random_list(min_size=2), cs.random_slice_indices())
# def test_subscript_heterogeneous_list_slice(input_list, slice):
#     """Test visitor of Subscript node representing slicing of heterogeneous list."""
#     assume(not isinstance(input_list[0], type(input_list[1])))
#     input_slice = ':'.join([str(index) if index else '' for index in slice])
#     program = f'{input_list}[{input_slice}]'
#     module, _ = cs._parse_text(program)
#     subscript_node = list(module.nodes_of_class(nodes.Subscript))[0]
#     assert subscript_node.inf_type.getValue() == List[Any]



import os
import sys
import webbrowser
from base64 import b64encode
from datetime import datetime
from http.server import BaseHTTPRequestHandler, HTTPServer

from jinja2 import Environment, FileSystemLoader
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import PythonLexer
from pylint.interfaces import IReporter
from pylint.reporters.ureports.nodes import BaseLayout

from .core import PythonTaReporter

TEMPLATES_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")


class HTMLReporter(PythonTaReporter):
    """Reporter that displays results in HTML form.

    By default, automatically opens the report in a web browser.
    """

    __implements__ = IReporter
    name = "HTMLReporter"

    _COLOURING = {
        "black": '<span class="black">',
        "black-line": '<span class="black line-num">',
        "bold": "<span>",
        "code-heading": "<span>",
        "style-heading": "<span>",
        "code-name": "<span>",
        "style-name": "<span>",
        "highlight": '<span class="highlight-pyta">',
        "grey": '<span class="grey">',
        "grey-line": '<span class="grey line-num">',
        "gbold": '<span class="gbold">',
        "gbold-line": '<span class="gbold line-num">',
        "reset": "</span>",
    }
    _PRE_LINE_NUM_SPACES = 0

    no_err_message = "No problems detected, good job!"
    no_snippet = "No code to display for this message."
    code_err_title = "Code Errors or Forbidden Usage (fix: high priority)"
    style_err_title = "Style or Convention Errors (fix: before submission)"
    OUTPUT_FILENAME = "pyta_report.html"

    def print_messages(self, level="all"):
        """Do nothing to print messages, since all are displayed in a single HTML file."""

    def display_messages(self, layout: BaseLayout) -> None:
        """Hook for displaying the messages of the reporter

        This will be called whenever the underlying messages
        needs to be displayed. For some reporters, it probably
        doesn't make sense to display messages as soon as they
        are available, so some mechanism of storing them could be used.
        This method can be implemented to display them after they've
        been aggregated.
        """
        grouped_messages = {path: self.group_messages(msgs) for path, msgs in self.messages.items()}

        template_f = self.linter.config.pyta_template_file
        template = Environment(loader=FileSystemLoader(TEMPLATES_DIR)).get_template(template_f)

        # Embed resources so the output html can go anywhere, independent of assets.
        # with open(os.path.join(TEMPLATES_DIR, 'pyta_logo_markdown.png'), 'rb+') as image_file:
        #     # Encode img binary to base64 (+33% size), decode to remove the "b'"
        #     pyta_logo_base64_encoded = b64encode(image_file.read()).decode()

        # Date/time (24 hour time) format:
        # Generated: ShortDay. ShortMonth. PaddedDay LongYear, Hour:Min:Sec
        dt = str(datetime.now().strftime("%a. %b. %d %Y, %I:%M:%S %p"))

        # Render the jinja template
        rendered_template = template.render(
            date_time=dt,
            reporter=self,
            grouped_messages=grouped_messages,
            os=os,
            enumerate=enumerate,
        )

        # If a filepath was specified, write to the file
        if self.out is not sys.stdout:
            self.writeln(rendered_template)
        else:
            rendered_template = rendered_template.encode("utf8")
            self._open_html_in_browser(rendered_template)

    def _open_html_in_browser(self, html: bytes) -> None:
        """
        Display html in a web browser without creating a temp file.
        Instantiates a trivial http server and uses the webbrowser module to
        open a URL to retrieve html from that server.

        Adapted from: https://github.com/plotly/plotly.py/blob/master/packages/python/plotly/plotly/io/_base_renderers.py#L655
        """

        class OneShotRequestHandler(BaseHTTPRequestHandler):
            def do_GET(self):
                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.end_headers()

                buffer_size = 1024 * 1024
                for i in range(0, len(html), buffer_size):
                    self.wfile.write(html[i : i + buffer_size])

            def log_message(self, format, *args):
                """Overridden so that no server logging is printed."""
                pass

        server = HTTPServer(("127.0.0.1", 0), OneShotRequestHandler)
        webbrowser.open(f"http://127.0.0.1:{server.server_port}", new=2)
        server.handle_request()
        server.server_close()
        print(
            "[INFO] Your PythonTA report is being opened in your web browser.\n"
            "       If it doesn't open, please add an output argument to python_ta.check_all\n"
            "       as follows:\n\n"
            "         check_all(..., output='pyta_report.html')\n\n"
            "       This will cause PythonTA to save the report to a file, pyta_report.html,\n"
            "       that you can open manually in a web browser.",
            file=sys.stderr,
        )

    @classmethod
    def _colourify(cls, colour_class: str, text: str) -> str:
        """Return a colourized version of text, using colour_class."""
        colour = cls._COLOURING[colour_class]
        new_text = text.lstrip(" ")
        space_count = len(text) - len(new_text)
        new_text = new_text.replace(" ", cls._SPACE)
        if "-line" not in colour_class:
            new_text = highlight(
                new_text,
                PythonLexer(),
                HtmlFormatter(nowrap=True, lineseparator="", classprefix="pygments-"),
            )

        return (space_count * cls._SPACE) + colour + new_text + cls._COLOURING["reset"]



from typing import *


class Monad:
    def __init__(self):
        raise NotImplementedError

    def bind(self, fn: Callable[[TypeVar("T")], "Monad"]) -> "Monad":
        raise NotImplementedError

    def __rshift__(self, fn: Callable[[TypeVar("T")], "Monad"]) -> "Monad":
        return self.bind(fn)

    def getValue(self):  # TODO: underscore naming
        return self.value


class Failable(Monad):
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.value == other.value

    def __str__(self):
        return self.value.__str__()

    def bind(self, fn):
        return fn(self.value)


def failable_map(fn: Callable[[TypeVar("T")], Failable], itr: Iterable[Failable]) -> Failable:
    lst = list(itr)
    if lst == []:
        return Failable([])
    return fn(lst[0]) >> (
        lambda fst: (failable_map(fn, lst[1:]) >> (lambda rest: Failable([fst] + rest)))
    )


def failable_collect(itr: Iterable[Failable]) -> Failable:
    return failable_map(lambda x: x, itr)



import os.path
from typing import Dict, Set

import graphviz
from astroid import nodes
from astroid.builder import AstroidBuilder

from python_ta.cfg import CFGBlock, CFGVisitor, ControlFlowGraph

USAGE = "USAGE: python -m sample_usage.draw_cfg <your-file.py>"
GRAPH_OPTIONS = {"format": "jpg", "node_attr": {"shape": "box", "fontname": "Courier New"}}


def display(cfgs: Dict[nodes.NodeNG, ControlFlowGraph], filename: str, view: bool = True) -> None:
    graph = graphviz.Digraph(name=filename, **GRAPH_OPTIONS)
    for node, cfg in cfgs.items():
        if isinstance(node, nodes.Module):
            subgraph_label = "__main__"
        elif isinstance(node, nodes.FunctionDef):
            subgraph_label = node.name
        else:
            continue
        with graph.subgraph(name=f"cluster_{id(node)}") as c:
            visited = set()
            _visit(cfg.start, c, visited)
            for block in cfg.unreachable_blocks:
                _visit(block, c, visited)
            c.attr(label=subgraph_label)

    graph.render(filename, view=view)


def _visit(block: CFGBlock, graph: graphviz.Digraph, visited: Set[int]) -> None:
    node_id = f"{graph.name}_{block.id}"
    if node_id in visited:
        return

    label = "\n".join([s.as_string() for s in block.statements]) + "\n"
    # Need to escape backslashes explicitly.
    label = label.replace("\\", "\\\\")
    # \l is used for left alignment.
    label = label.replace("\n", "\\l")

    fill_color = "grey93" if not block.reachable else "white"

    graph.node(node_id, label=label, fillcolor=fill_color, style="filled")
    visited.add(node_id)

    for edge in block.successors:
        graph.edge(node_id, f"{graph.name}_{edge.target.id}")
        _visit(edge.target, graph, visited)


def main(filepath: str) -> None:
    filename = os.path.splitext(os.path.basename(filepath))[0]
    mod = AstroidBuilder().file_build(filepath)
    visitor = CFGVisitor()
    mod.accept(visitor)

    display(visitor.cfgs, filename)


if __name__ == "__main__":
    import sys

    if len(sys.argv) < 2:
        print(USAGE)
        exit(1)

    filepath = sys.argv[1]
    main(filepath)



"""checker for variables that might not be defined in the program.
"""
from itertools import chain
from typing import Generator, Set, Union

from astroid import nodes
from pylint.checkers import BaseChecker, utils
from pylint.checkers.utils import check_messages
from pylint.interfaces import IAstroidChecker

from python_ta.cfg.graph import CFGBlock, ControlFlowGraph


class PossiblyUndefinedChecker(BaseChecker):

    __implements__ = IAstroidChecker
    # name is the same as file name but without _checker part
    name = "possibly_undefined"
    # use dashes for connecting words in message symbol
    msgs = {
        "E9969": (
            "This variable might not be defined when this statement is executed.",
            "possibly-undefined",
            "Reported when a statement uses a variable that might not be assigned.",
        )
    }

    # this is important so that your checker is executed before others
    priority = -1

    def __init__(self, linter=None):
        super().__init__(linter=linter)
        self._possibly_undefined: Set[nodes.Name] = set()

    @check_messages("possibly-undefined")
    def visit_name(self, node):
        """Adds message if there exists a path from the start block to node where
        a variable used by node might not be defined."""
        if node in self._possibly_undefined:
            self.add_message("possibly-undefined", node=node)

    def visit_module(self, node):
        self._analyze(node)

    def visit_functiondef(self, node):
        self._analyze(node)

    def _analyze(self, node: Union[nodes.Module, nodes.FunctionDef]) -> None:
        """Runs the data flow algorithm on a `Module` or `Function` CFG, which in turn
        appends `Name` nodes to `self.possibly_undefined` if it might not be defined.

        Data flow algorithms retrieved from:
        https://www.seas.harvard.edu/courses/cs252/2011sp/slides/Lec02-Dataflow.pdf#page=31
        """
        out_facts = {}
        cfg = ControlFlowGraph()
        cfg.start = node.cfg_block
        blocks = list(cfg.get_blocks_postorder())
        blocks.reverse()

        all_assigns = self._get_assigns(node)
        for block in blocks:
            out_facts[block] = all_assigns.copy()

        worklist = blocks
        while len(worklist) != 0:
            b = worklist.pop()
            outs = [out_facts[p.source] for p in b.predecessors if p.source in out_facts]
            if outs == []:
                in_facts = set()
            else:
                in_facts = set.intersection(*outs)
            temp = self._transfer(b, in_facts, all_assigns)
            if temp != out_facts[b]:
                out_facts[b] = temp
                worklist.extend([succ.target for succ in b.successors])

    def _transfer(self, block: CFGBlock, in_facts: Set[str], local_vars: Set[str]) -> Set[str]:
        gen = in_facts.copy()
        kill = set()
        for statement in block.statements:
            if isinstance(statement, nodes.FunctionDef):
                continue
            for node in self.get_nodes(statement):
                if isinstance(node, nodes.AssignName):
                    gen.add(node.name)
                elif isinstance(node, nodes.DelName):
                    kill.add(node.name)
                else:
                    name = node.name
                    # comment out 'self.config....' check when running tests
                    if (
                        not (name in nodes.Module.scope_attrs or utils.is_builtin(name))
                        and name in local_vars
                        and name not in gen.difference(kill)
                    ):
                        self._possibly_undefined.add(node)
                    elif node in self._possibly_undefined:
                        self._possibly_undefined.remove(node)
        return gen.difference(kill)

    def _get_assigns(self, node: Union[nodes.FunctionDef, nodes.Module]) -> Set[str]:
        """Returns a set of all local and parameter variables that could be
        defined in the program (either a function or module).

        IF a variable 'v' is defined in a function and there is no global/nonlocal
        statement applied to 'v' THEN 'v' is a local variable.

        Note that `local variable` in the context of a module level analysis,
        refers to global variables.
        """
        assigns = set()
        kills = set()
        for name, assign_nodes in node.scope().locals.items():
            if any(isinstance(elem, nodes.AssignName) for elem in assign_nodes):
                assigns.add(name)
        for statement in node.nodes_of_class(
            (nodes.Nonlocal, nodes.Global, nodes.ImportFrom, nodes.Import),
            nodes.FunctionDef,
        ):
            for name in statement.names:
                if type(name) is tuple:
                    # name[1] is the alias of the imported object/var name[0]
                    # name[1] == str or None
                    name = name[1] or name[0]
                kills.add(name)

        return assigns.difference(kills)

    def get_nodes(self, statement: nodes.NodeNG) -> Generator[nodes.NodeNG, None, None]:
        multiple_nodes = lambda nodes_: chain.from_iterable(self.get_nodes(node) for node in nodes_)
        if isinstance(statement, nodes.Assign):
            # RHS is evaluated before assigned in an assignment statement
            yield from self.get_nodes(statement.value)
            yield from multiple_nodes(statement.targets)  # statement.targets is a list of nodes
        elif isinstance(
            statement, (nodes.ListComp, nodes.SetComp, nodes.DictComp, nodes.GeneratorExp)
        ):
            # Comprehension targets are assigned before expression is evaluated.
            yield from multiple_nodes(
                statement.generators
            )  # statement.generators is a list of nodes
            if not hasattr(statement, "elt"):
                yield from self.get_nodes(statement.key)  # keys evaluated first
                yield from self.get_nodes(statement.value)
            else:
                yield from self.get_nodes(statement.elt)
        else:
            yield from statement.nodes_of_class(
                (nodes.AssignName, nodes.DelName, nodes.Name), nodes.FunctionDef
            )


def register(linter):
    linter.register_checker(PossiblyUndefinedChecker(linter))



from typing import *

from astroid import nodes
from pytest import skip

from python_ta.typecheck.base import TypeFail, TypeFailStarred

from .. import custom_hypothesis_support as cs


def test_list():
    src = """
    *a, b = [1, 2, 3]
    a
    b
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for name_node in ast_mod.nodes_of_class(nodes.Name):
        assert not isinstance(name_node.inf_type, TypeFail)
        if name_node.name == "a":
            assert name_node.inf_type.getValue() == List[int]
        elif name_node.name == "b":
            assert name_node.inf_type.getValue() == int


def test_range():
    src = """
    *a, b = range(5)
    a
    b
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for name_node in ast_mod.nodes_of_class(nodes.Name):
        assert not isinstance(name_node.inf_type, TypeFail)
        if name_node.name == "a":
            t = name_node.inf_type.getValue()
            assert ti.type_constraints.resolve(t).getValue() == List[int]
        elif name_node.name == "b":
            assert name_node.inf_type.getValue() == int


def test_tuple():
    src = """
    *a, b = (1, 2, 3)
    a
    b
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for name_node in ast_mod.nodes_of_class(nodes.Name):
        assert not isinstance(name_node.inf_type, TypeFail)
        if name_node.name == "a":
            assert name_node.inf_type.getValue() == List[int]
        elif name_node.name == "b":
            assert name_node.inf_type.getValue() == int


def test_order():
    src = """
    a, *b, c = [1, 2, 3, 4]
    a
    b
    c
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for name_node in ast_mod.nodes_of_class(nodes.Name):
        assert not isinstance(name_node.inf_type, TypeFail)
        if name_node.name == "a":
            assert name_node.inf_type.getValue() == int
        elif name_node.name == "b":
            assert name_node.inf_type.getValue() == List[int]
        elif name_node.name == "c":
            assert name_node.inf_type.getValue() == int


def test_order_2():
    src = """
    a, *b = [1, 2, 3]
    a
    b
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for name_node in ast_mod.nodes_of_class(nodes.Name):
        assert not isinstance(name_node.inf_type, TypeFail)
        if name_node.name == "a":
            assert name_node.inf_type.getValue() == int
        elif name_node.name == "b":
            assert name_node.inf_type.getValue() == List[int]


def test_mixed_tuple():
    src = """
    *a, b = (1, "Hello", False)
    a
    b
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for name_node in ast_mod.nodes_of_class(nodes.Name):
        assert not isinstance(name_node.inf_type, TypeFail)
        if name_node.name == "a":
            assert name_node.inf_type.getValue() == List[Any]
        elif name_node.name == "b":
            assert name_node.inf_type.getValue() == bool


def test_mixed_tuple_order():
    src = """
    a, *b = (1, "Hello", False)
    a
    b
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for name_node in ast_mod.nodes_of_class(nodes.Name):
        assert not isinstance(name_node.inf_type, TypeFail)
        if name_node.name == "a":
            assert name_node.inf_type.getValue() == int
        elif name_node.name == "b":
            assert name_node.inf_type.getValue() == List[Any]


def test_mixed_tuple_three_var():
    src = """
    a, *b, c = (1, "Hello", False, "Goodbye", 4, "What")
    a
    b
    c
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for name_node in ast_mod.nodes_of_class(nodes.Name):
        assert not isinstance(name_node.inf_type, TypeFail)
        if name_node.name == "a":
            assert name_node.inf_type.getValue() == int
        elif name_node.name == "b":
            assert name_node.inf_type.getValue() == List[Any]
        elif name_node.name == "c":
            assert name_node.inf_type.getValue() == str


def test_mixed_tuple_four_var():
    src = """
    a, *b, c, d = (1, "Hello", "Good morning", "Goodbye", 4, "What")
    a
    b
    c
    d
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for name_node in ast_mod.nodes_of_class(nodes.Name):
        assert not isinstance(name_node.inf_type, TypeFail)
        if name_node.name == "a":
            assert name_node.inf_type.getValue() == int
        elif name_node.name == "b":
            assert name_node.inf_type.getValue() == List[str]
        elif name_node.name == "c":
            assert name_node.inf_type.getValue() == int
        elif name_node.name == "d":
            assert name_node.inf_type.getValue() == str


def test_multi_starred():
    src = """
    *a, b, *c = [1, 2, 3, 4, 5]
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    assign_node = next(ast_mod.nodes_of_class(nodes.Assign))
    assert isinstance(assign_node.inf_type, TypeFailStarred)


def test_multi_variable():
    src = """
    lst = [1, 2, 3, 4, 5]
    *a, b, *c = lst
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    assign_node = list(ast_mod.nodes_of_class(nodes.Assign))[1]
    assert isinstance(assign_node.inf_type, TypeFailStarred)



import json
from typing import Dict, List

from pylint.interfaces import IReporter
from pylint.reporters.ureports.nodes import BaseLayout

from .core import NewMessage, PythonTaReporter


class JSONReporter(PythonTaReporter):
    """Reporter that outputs JSON.

    Based on Pylint's JSONReporter.
    """

    __implements__ = IReporter
    name = "JSONReporter"

    OUTPUT_FILENAME = "pyta_report.json"

    messages: Dict[str, List[NewMessage]]

    def display_messages(self, layout: BaseLayout) -> None:
        """Hook for displaying the messages of the reporter

        This will be called whenever the underlying messages
        needs to be displayed. For some reporters, it probably
        doesn't make sense to display messages as soon as they
        are available, so some mechanism of storing them could be used.
        This method can be implemented to display them after they've
        been aggregated.
        """
        output = []
        for k, msgs in self.messages.items():
            output.append(
                {
                    "filename": k,
                    "msgs": [
                        {**msg._replace(node=None)._asdict(), **self._snippet_endings(msg)}
                        for msg in msgs
                    ],
                }
            )

        self.writeln(json.dumps(output, indent=4))

    def _snippet_endings(self, msg: NewMessage) -> Dict[str, int]:
        """Return a dictionary containing the index of the last line and the index of the last column of the
        error reported by the message in msg.
        """
        if msg.node is None:
            line_end, column_end = msg.line, len(self.source_lines[msg.line - 1])
        else:
            line_end, column_end = msg.node.end_lineno, msg.node.end_col_offset
        return {"line_end": line_end, "column_end": column_end}



from typing import List

import pytest

from python_ta.contracts import check_all_contracts


class Employee:
    """
    Represents an employee

    Representation Invariants:
    - len(self.name) > 0
    - self.wage >= 15
    """

    name: str
    wage: float

    def __init__(self, name, wage):
        self.name = name
        self.wage = wage

    def change_wages(self, new_wage):
        """
        Precondition: self.wage < new_wage
        """
        self.wage = new_wage


class Developer(Employee):
    """
    Represents a sofware developer

    Representation Invariants:
    - self.preferred_language == "Python" or self.preferred_language == "Java"
    """

    preferred_language: str

    def __init__(self, name, wage, preferred_language):
        Employee.__init__(self, name, wage)
        self.preferred_language = preferred_language


class Teacher(Employee):
    """
    Represents a restaurant teacher

    Representation Invariants:
    - self.wage == self.wage_per_class * len(self.currently_teaching)
    """

    currently_teaching: List[str]
    wage_per_class: float

    def __init__(self, name, wage_per_class, currently_teaching):
        Employee.__init__(self, name, wage_per_class * len(currently_teaching))
        self.wage_per_class = wage_per_class
        self.currently_teaching = currently_teaching

    def teach_new_classes(self, currently_teaching):
        # temporary violation of Teacher rep invariant (if before/after lengths are different)
        self.change_wages(self.wage_per_class * len(currently_teaching))

        # amending violation
        self.currently_teaching = currently_teaching

    def update_wage_per_class(self, wage_per_class):
        # temporary violation of Teacher rep invariant (if wage rate different from before)
        self.wage_per_class = wage_per_class

        # amending violation
        self.change_wages(wage_per_class * len(self.currently_teaching))


class TeamMember:
    """
    Represents a person on a team

    Representation Invariants:
    - len(self.team) > 0
    """

    team: str

    def __init__(self, team):
        self.team = team


class TeamLead(Developer, TeamMember):
    """
    Represents a team lead

    Representation Invariants:
    - self.wage >= 30
    """

    def __init__(self, name, wage, preferred_language, team):
        Developer.__init__(self, name, wage, preferred_language)
        TeamMember.__init__(self, team)


# Decorating everything in this file
check_all_contracts(__name__, decorate_main=False)


@pytest.fixture
def developer():
    return Developer("Ibrahim", 35, "Python")


@pytest.fixture
def teamlead():
    return TeamLead("David", 50, "Python", "PyTA")


@pytest.fixture
def pe_bio_teacher():
    return Teacher("Jewel", 25, ["P.E.", "Biology"])


def test_change_developer_wage_lower(developer) -> None:
    """
    Change the wage to a lower amount, expect an exception.
    """
    with pytest.raises(AssertionError) as excinfo:
        developer.change_wages(25)
    msg = str(excinfo.value)
    assert "self.wage < new_wage" in msg


def test_increase_teamlead_wages(teamlead):
    """
    Increases the wages of the team lead.
    """
    teamlead.change_wages(75)
    assert teamlead.wage == 75


def test_decrease_teamlead_wages(teamlead):
    """
    Decrease the wages of the team lead to below 30. Expect an exception.
    """
    with pytest.raises(AssertionError) as excinfo:
        teamlead.wage = 20
    msg = str(excinfo.value)
    assert "self.wage >= 30" in msg


def test_change_teamlead_language(teamlead):
    """
    Change the preferred language to a wrong value. Expect an exception.
    """
    with pytest.raises(AssertionError) as excinfo:
        teamlead.preferred_language = "C++"
    msg = str(excinfo.value)
    assert 'self.preferred_language == "Python" or self.preferred_language == "Java"' in msg


def test_change_teamlead_name(teamlead):
    """
    Change the name of the teamlead.
    """
    teamlead.name = "Ignas"
    assert teamlead.name == "Ignas"


def test_change_teamlead_team_invalid(teamlead):
    """
    Changes the team of the teamlead to something invalid. Excpects an exception.
    """
    with pytest.raises(AssertionError) as excinfo:
        teamlead.team = ""
    msg = str(excinfo.value)
    assert "len(self.team) > 0" in msg


def test_call_super_when_temp_invalid(pe_bio_teacher):
    """
    Temporarily violates a rep invariant by setting an instance attribute but later remedies the
    violation by calling a method defined in the super class. Expects no exception.
    """
    pe_bio_teacher.update_wage_per_class(100)
    assert pe_bio_teacher.wage_per_class == 100
    assert pe_bio_teacher.wage == pe_bio_teacher.wage_per_class * len(
        pe_bio_teacher.currently_teaching
    )


def test_call_super_creates_temp_invalid(pe_bio_teacher):
    """
    Call an instance method that temporarily violates a rep invariant when calling a method
    defined in a super class but later remedies the violation. Expects no exception.
    """
    pe_bio_teacher.teach_new_classes(["Computer Science", "Video Production", "Life Science"])
    assert len(pe_bio_teacher.currently_teaching) == 3
    assert pe_bio_teacher.wage == pe_bio_teacher.wage_per_class * len(
        pe_bio_teacher.currently_teaching
    )



from __future__ import annotations

from typing import *
from typing import _GenericAlias

from astroid import nodes

from python_ta.utils import _get_name, _gorg

###############################################################################
# Operator translations into dunder method name and English name.
###############################################################################
UNARY_TO_ENGLISH = {"+": "apply unary + to", "-": "negate", "~": "take the bitwise inverse of"}

UNARY_TO_METHOD = {"+": "__pos__", "-": "__neg__", "~": "__invert__"}

BINOP_TO_ENGLISH = {
    "+": "add",
    "-": "subtract",
    "*": "multiply",
    "//": "use integer division with",
    "%": "use modulus with",
    "/": "use floating-point division with",
    "**": "exponentiate",
    "&": "use bitwise AND with",
    "^": "use bitwise XOR with",
    "|": "use bitwise OR with",
    "<<": "apply a bitshift to the left",
    ">>": "apply a bitshift to the right",
    "==": "compare",
    "!=": "compare",
    "<": "compare",
    "<=": "compare",
    ">": "compare",
    ">=": "compare",
    # TODO : 'is' and 'in'
}

BINOP_TO_METHOD = {
    "+": "__add__",
    "+=": "__iadd__",
    "-": "__sub__",
    "-=": "__isub__",
    "*": "__mul__",
    "*=": "__imul__",
    "//": "__floordiv__",
    "//=": "__ifloordiv__",
    "%": "__mod__",
    "%=": "__imod__",
    "/": "__truediv__",
    "**": "__pow__",
    "**=": "__ipow__",
    "&": "__and__",
    "&=": "__iand__",
    "^": "__xor__",
    "^=": "__ixor__",
    "|": "__or__",
    "|=": "__ior__",
    "<<": "__lshift__",
    "<<=": "__ilshift__",
    ">>": "__rshift__",
    ">>==": "__irshift__",
    "==": "__eq__",
    "!=": "__ne__",
    "<": "__lt__",
    "<=": "__le__",
    ">": "__gt__",
    ">=": "__ge__",
    "in": "__contains__",
    "not in": "__contains__",
}

BINOP_TO_REV_METHOD = {
    "+": "__radd__",
    "-": "__rsub__",
    "*": "__rmul__",
    "//": "__rfloordiv__",
    "%": "__rmod__",
    "/": "__rtruediv__",
    "**": "__rpow__",
    "&": "__rand__",
    "^": "__rxor__",
    "|": "__ror__",
    "<<": "__rlshift__",
    ">>": "__rrshift__",
}

INPLACE_TO_BINOP = {
    "+=": "+",
    "-=": "-",
    "*=": "*",
    "//=": "//",
    "%=": "%",
    "**=": "*",
    "&=": "&",
    "^=": "^",
    "|=": "=",
    "<<=": "<<",
    ">>=": ">>",
}


###############################################################################
# Error message
###############################################################################
def error_message(tf: TypeFail) -> None:
    """Return an appropriate error message given an instance of TypeFailFunction."""
    if isinstance(tf.src_node, nodes.UnaryOp):
        return unaryop_error_message(tf.src_node)
    elif isinstance(tf.src_node, nodes.BinOp):
        return binop_error_message(tf.src_node)
    elif isinstance(tf.src_node, nodes.Subscript):
        return subscript_error_message(tf.src_node)
    else:
        return f"TypeFail: Invalid function call at {tf.src_node.as_string()}"


###############################################################################
# BinOp message
###############################################################################
# TODO: Convert this into dictionary
def binary_op_hints(op, args):
    """Return an appropriate 'hint' or suggestion given the binary operation and operand types."""
    if op == "+":
        if "int" in args and "str" in args:
            return "Perhaps you wanted to cast the integer into a string or vice versa?"


def binop_error_message(node: nodes.BinOp) -> str:
    op_name = BINOP_TO_ENGLISH[node.op]
    left_type = _get_name(node.left.inf_type.getValue())
    right_type = _get_name(node.right.inf_type.getValue())
    hint = binary_op_hints(node.op, [left_type, right_type]) or ""

    return (
        f"You cannot {op_name} {_correct_article(left_type)}, {node.left.as_string()}, "
        f"and {_correct_article(right_type)}, {node.right.as_string()}. "
        f"{hint}"
    )


###############################################################################
# UnaryOp message
###############################################################################
def unaryop_error_message(node: nodes.UnaryOp) -> str:
    op_name = UNARY_TO_ENGLISH[node.op]
    operand = _get_name(node.operand.inf_type.getValue())

    return f"You cannot {op_name} {_correct_article(operand)}, {node.operand.as_string()}."


###############################################################################
# Subscript message
###############################################################################
def subscript_error_message(node: nodes.Subscript) -> str:
    # Accessing an element of a List with an incompatible index type (non-integers)
    subscript_concrete_type = (node.value.inf_type).getValue()
    if subscript_concrete_type is type(None):
        return f"NoneType is not subscriptable."

    if not isinstance(subscript_concrete_type, _GenericAlias):
        subscript_gorg = subscript_concrete_type
    else:
        subscript_gorg = _gorg(subscript_concrete_type)

    if subscript_gorg is list:
        slice_type = _get_name(node.slice.inf_type.getValue())
        slice_val = node.slice.value
        slice_str = slice_val.as_string() if isinstance(slice_val, nodes.NodeNG) else str(slice_val)
        return (
            f"You can only access elements of a list using an int. "
            f"You used {_correct_article(slice_type)}, {slice_str}."
        )
    elif subscript_gorg is tuple:
        slice_type = _get_name(node.slice.inf_type.getValue())
        slice_val = node.slice.value
        slice_str = slice_val.as_string() if isinstance(slice_val, nodes.NodeNG) else str(slice_val)
        return (
            f"You can only access elements of a tuple using an int. "
            f"You used {_correct_article(slice_type)}, {slice_str}."
        )
    elif subscript_gorg is dict:
        slice_type = _get_name(node.slice.inf_type.getValue())
        slice_val = node.slice.value
        slice_str = slice_val.as_string() if isinstance(slice_val, nodes.NodeNG) else str(slice_val)
        return (
            f"You tried to access an element of this dictionary using "
            f"{_correct_article(slice_type)}, {slice_str}, "
            f"but the keys are of type {_get_name(subscript_concrete_type.__args__[0])}."
        )
    else:
        return f"You make a type annotation with an incorrect subscript."


def _correct_article(noun: str) -> str:
    """Helper to return a noun with the correct article."""
    if noun.lower()[0] in "aeiou":
        return "an " + noun
    else:
        return "a " + noun



"""Patch pylint error messages."""
from importlib import import_module

# global
patch_data = {
    "pylint.checkers.base": {
        "BasicChecker": {
            "E0111": "reversed() can only be called on sequence types like a str, list, or tuple.",
            "W0101": "Code after a return or raise statement will never be run, so you should "
            "either remove this code or review the above return/raise statement(s).",
            "W0102": "Mutable values (like %s) should not be used as a default arguments for functions. "
            "This is a common cause of confusing and unwanted behaviour.",
            "W0104": "This statement doesn't have any effect, and could be "
            "removed without changing the behaviour of the program.",
            "W0106": "Right now, expression %s isn't being used by the rest "
            "of your code. You should either remove this expression or "
            "save it in a variable to use it later.",
            "W0108": "This lambda function isn't necessary since you can pass "
            "the arguments directly to the function you're calling in "
            "the lambda's body.",
            "W0109": "Duplicate key %r in dictionary literal.",
            "W0122": "Using exec() to evaluate code represented as a string "
            "can be potentially dangerous, and is typically a sign of poor design.",
            "W0123": "Using eval() to evaluate code represented as a string "
            "can be potentially dangerous, and is typically a sign of poor design.",
            "W0126": "This if condition references a function but is missing the parentheses "
            "to call the function. This will cause the condition to always evaluate to True.",
            "W0199": "Calling assert on a tuple will only evaluate to False if "
            "it is empty. It's likely you intended 'assert x, y' rather "
            "than 'assert (x,y)'.",
        },
        "ComparisonChecker": {
            "R0123": "When performing a comparison with a literal, use '==' instead of 'is'. "
            "Exceptions: use 'is' when comparing to None.",
            "C0123": "You should use 'isinstance(x, my_type)' instead of 'type(x) == my_type' to "
            "check the type of a value.",
        },
        "NameChecker": {
            "C0104": 'Disallowed name "%s". You should give your variables '
            "meaningful rather than generic names.",
            "C0103": '%s name "%s" doesn\'t conform to %s. '
            "Please refer to the PyTA Help Documentation for more information "
            "on Python naming conventions.",
        },
        "PassChecker": {"W0107": "Unnecessary pass statement (you can just remove this)."},
        "BasicErrorChecker": {
            "E0101": "__init__ is called only to initialize the attributes of a new object. "
            "It should not return anything.",
            "E0102": "%s was already defined line %s. "
            "You should make sure all functions, methods and classes "
            "defined have different names.",
            "E0103": "The %r keyword should only be used inside a loop.",
            "E0104": "The 'return' keyword is used to end the execution of a function "
            "and return a result, so you should only use it inside of a function body.",
            "E0108": "Duplicate parameter name %s in function definition.",
        },
    },
    "pylint.checkers.refactoring": {
        "RefactoringChecker": {
            "R1701": "These multiple calls to isinstance can be merged into isinstance(%s, (%s)).",
            "R1702": "Your code has too many nested blocks (%s, exceeding limit %s). "
            "You should try to break it down into helper functions.",
            "R1704": "This local variable %r should have a name different from all function parameters.",
            "R1707": "If you mean to create a tuple here, surround the tuple expression with parentheses.",
            "R1710": "Since this function can return a non-None value, you should "
            'explicitly write "return None" whenever None is returned by '
            "this function. (Possibly including at the end of the function body.)",
            "R1712": "Use parallel assignment to swap the values of variables. "
            "For example: to swap variables 'a' and 'b', simply do 'a, b = b, a'.",
            "R1713": "If you want to concatenate a sequence of strings, use the str.join method.",
            "R1714": "To check whether this variable is equal to one of many values, merge the "
            "comparisons into a single check: %r.",
            "R1715": "To lookup a value from a dictionary if a key is present and "
            "use a default value if the key is not, use the built-in dict.get method.",
            "R1716": "You can simplify this boolean operation into a chained comparison. "
            "Instead of 'a < b and b < c', use 'a < b < c'.",
            "R1721": "When you want to convert a collection from one type into another, use the appropriate "
            "type constructor instead of a comprehension. Here, use %s instead.",
        },
        "NotChecker": {
            "C0113": 'Change "%s" to "%s". Removing the negation simplifies the expression.'
        },
        "RecommendationChecker": {
            "C0201": "Rather than using '.keys()' to iterate over the keys of a dictionary, "
            "you can simply iterate over the dictionary itself."
        },
        "LenChecker": {
            "C1801": "When checking if a collection is empty, either check equality with an empty instance "
            "of the collection type (e.g., == []) or explicitly check that the len is equal to 0."
        },
    },
    "pylint.checkers.design_analysis": {
        "MisdesignChecker": {
            "R0912": "This function has too many branches (%s, exceeding limit %s). "
            "You should try to reduce the complexity of the function "
            "by splitting it up into helper functions.",
            "R0913": "This function has too many parameters (%s, exceeding limit %s). "
            "You should try to reduce the complexity of the function "
            "by splitting up it, or combining related objects as a single one.",
            "R0914": "This function has too many local variables (%s, exceeding limit %s). "
            "You should try to reduce the complexity of the function "
            "by splitting up it, or combining related objects as a single one.",
            "R0915": "This function has too many statements ((%s, exceeding limit %s). "
            "It should be split up into smaller functions.",
            "R0916": "This if condition contains too many boolean expressions "
            "(%s, exceeding limit %s). You should see if you can reduce "
            "the amount of expressions, or split it into branches.",
        }
    },
    "pylint.checkers.classes": {
        "ClassChecker": {
            "E0203": "You can't use instance attribute %r prior to its assignment " "on line %s.",
            "E0211": "Every instance method requires at least one parameter 'self', "
            "referring to the object on which this method is called. ",
            "E0213": "The first parameter of a method should always be called 'self'. ",
            "E0241": "This class %r should not inherit from the same class multiple times.",
            "R0201": "This method does not make use of 'self', so you should remove the self "
            "parameter and move this method outside of the class (turning it into a "
            "top-level function).",
            "R0205": "All classes inherit from object by default, so you do not need to specify "
            "this in the header for %r.",
            "W0211": "The static method %r should not have 'self' as the first parameter.",
            "W0212": "Since %s starts with an underscore, it is considered private and so "
            "should not be accessed outside of the class in which it is defined.",
            "W0222": "This method's parameters must have the same name, order, and default "
            "arguments as %s %r method.",
            "W0223": "The abstract method %r in class %r must be overridden within a "
            "concrete subclass.",
            "W0231": "Subclass '__init__' method should call the '__init__' method from "
            "its superclass %r.",
            "W0233": "Subclass '__init__' method should call the '__init__' method of "
            "the superclass rather than some unrelated class %r.",
        },
        "SpecialMethodsChecker": {
            "E0301": "An '__iter__' method must return an iterator, i.e an object with a "
            "'__next__' method.",
            "E0302": "The special method %r expects to take %s parameters, but %d %s given.",
        },
    },
    "pylint.checkers.exceptions": {
        "ExceptionsChecker": {
            "E0701": "'except' blocks are checked top to bottom, so generic exception types should "
            "come after specific exception types. Otherwise, the block for the specific "
            "exception type will never be reached. (%s)",
            "E0702": "'raise' is used to raise exceptions. Anything else will lead to an error. "
            "(Raising %s)",
            "E0704": "This 'raise' statement must have an exception class or instance.",
            "E0710": "This 'raise' statement is using a class that is not a subclass of Exception.",
            "E0711": "You mean to raise 'NotImplementedError', which indicates that a method is abstract. ",
            "E0712": "'except' expects a class that is a subclass of the Exception class (%s is not).",
            "W0702": "It's bad practice to use the 'except' keyword without passing an exception, "
            "since any and all expressions will be caught, which may lead to undetected errors "
            "in your code.",
            "W0703": "It's bad practice to use 'except: %s' since any and all expressions "
            "will be caught, which may lead to undetected errors in your code.",
            "W0705": "'except' blocks are checked top to bottom, so if we try to catch the same "
            "exception %s multiple times, only the first 'except' block for the exception "
            "will be reached.",
            "W0706": "If the 'except' block uses 'raise' as its first or only statement, it "
            "will just raise back the exception immediately.",
            "W0711": "'except' can handle multiple exceptions, but they have to be written as a "
            "tuple. For example, use 'except (A, B)' instead of 'except A %s B'",
            "W0716": "This is not a valid operation on exceptions. %s",
        }
    },
    "pylint.checkers.format": {
        "FormatChecker": {
            "W0301": "You should remove this semicolon. In Python, statements are separated by "
            "starting each one on a new line.",
            "C0301": "This line is %s characters long, exceeding the limit of %s characters.",
            "C0303": "Trailing whitespace is poor formatting, and should be removed.",
            "C0304": "By convention, each Python file should end with a blank line. To fix this, "
            "go down to the bottom of your file and press Enter/Return to add an empty line.",
            "C0321": "Avoid writing multiple statements on a single line. "
            "This makes your code harder to understand.",
        }
    },
    "pylint.checkers.imports": {
        "ImportsChecker": {
            "E0401": "Unable to import %s. Check the spelling of the module name, or "
            "whether the module is installed (if it is a module you did not write), "
            "or whether the module is in the same directory as this file "
            "(if it is a module you did write).",
            "W0401": "Wildcard (*) imports add all objects from the imported module as "
            "global variables, which may result in name collisions. You should only "
            "import from %s what you need, or import the whole module "
            '("import module") and use dot notation to access the module\'s functions/classes.',
            "W0404": "Module %r was re-imported on line %s. A module should not be "
            "imported more than once.",
            "W0410": "Imports of '__future__' must be the first non-docstring statement in the module.",
            "C0410": "Import each module on line %s on separately, one per line.",
            "C0414": "The 'import as' is used to import a module and give it a different name in this file. "
            "If you don't want to rename the module, simply use 'import' instead.",
        }
    },
    "pylint.checkers.newstyle": {
        "NewStyleConflictChecker": {
            "E1003": "The first argument to 'super' should be the current class. "
            "The argument %r is not valid."
        }
    },
    "pylint.checkers.stdlib": {
        "StdlibChecker": {
            "W1501": "\"%s\" is not a valid mode for opening a file. Common modes are 'r' for reading a file "
            "and 'w' for writing to a file."
        }
    },
    "pylint.checkers.strings": {
        "StringFormatChecker": {
            "E1307": "The argument %r does not match type %r specified by the format string.",
            "W1309": 'This f-string does not have any {...} expressions, and so the "f" can be removed.',
            "E1310": "Suspicious argument in %s.%s call. Strip will remove any characters contained "
            "in its argument, so duplicate characters are either unintended, or redundant.",
        },
        "StringConstantChecker": {
            "W1401": "The string '%s' contains a backslash, but is not part of an escape sequence. "
            "If this is not a typo, then you should make this explicit by escaping the "
            "backslash with another backslash.",
            "W1402": "Unicode escape being used in byte string '%s'. Since this is not a "
            "unicode string, inserting unicode characters will have no effect. If "
            "you just want to write '\\u', you should make this explicit by escaping "
            "the backslash with another backslash.",
            "W1404": "This %s has two consecutive string literals. You are likely missing a "
            "comma between them.",
        },
    },
    "pylint.checkers.typecheck": {
        "TypeChecker": {
            "E1101": "%s %r doesn't have the method/attribute %r%s.",
            "E1102": "%s is not callable. You may have written unnecessary parentheses following a variable.",
            "E1126": "Objects can only be indexed by ints, slices, or objects with an '__index__' method.",
            "E1127": "The index in a slice must be an int, 'None', or an object with an "
            "'__index__' method.",
            "E1129": "The object after the 'with' keyword must be an instance of a type that implements the "
            "'__enter__' and '__exit__', methods, which type %s does not.",
            "E1131": "%s. The binary operation can't be performed on the two operands provided.",
            "E1136": "%s does not support indexing.",
            "E1140": "Dictionary keys must be immutable.",
            "E1141": "If you want to iterate through the (key, value) pairs of a dictionary, you need to write "
            "'for key, value in my_dict.items()'.",
            "W1114": "The arguments in this function call aren't in the same order as the parameters in the function "
            "definition.",
        },
        "IterableChecker": {
            "E1133": "Can not iterate over %s as it is not an instance of an iterable type."
        },
    },
    "pylint.checkers.variables": {
        "VariablesChecker": {
            "E0601": "The variable %r may be accessed prior to it having a value assigned.",
            "E0611": "Can't find the name %r to import from module %r.",
            "E0633": "Trying to assign multiple values from%s, but it is not a sequence.",
            "W0604": "Every global variable can be referenced from the module level, so using "
            "the 'global' keyword at the module level has no effect.",
            "W0611": "The import %s is unused, and so can be removed.",
            "W0612": "The variable %s is unused and can be removed. If you intended to use it, "
            "there may be a typo elsewhere in the code.",
            "W0613": "The parameter %r is unused. This may indicate you misspelled a parameter name "
            "in the function body. Otherwise, the parameter can be removed from the function "
            "without altering the program.",
            "W0621": "The variable name %r already refers to a variable in the outer scope defined on "
            "line %s. This makes your code harder to read due to having to keep track of "
            "which variable we're referencing.",
            "W0622": "%r is the name of a built-in function, and should not be overridden.",
            "W0631": "Variables defined by a loop are only accessible within the loop, meaning "
            "the variable %r is possibly undefined.",
            "W0632": "The sequence%s is being used to assign %d variables but contains %d values.",
            "W0642": "%s is a local variable within the function. Reassigning this variable "
            "doesn't mutate any objects, and so has no effect after the function ends.",
        }
    },
}


# We are assuming only the first elements of the tuple values in <msgs> are being patched.
def patch_error_messages():
    """Patch <msgs> in pylint checkers to make them more helpful for python_ta clients."""
    for file_name, file_data in patch_data.items():
        for checker_name, checker_data in file_data.items():
            # only access the specific variable being changed
            checker = getattr(import_module(file_name), checker_name)
            if hasattr(checker, "msgs"):
                for error_id, new_msg in checker_data.items():
                    lst_msg = list(checker.msgs[error_id])
                    # first element of the tuple value changed
                    lst_msg[0] = new_msg
                    checker.msgs[error_id] = tuple(lst_msg)
            else:
                print("no msgs attribute!")



import os
from collections import defaultdict
from typing import *
from typing import Any, ForwardRef

from astroid import nodes
from astroid.builder import AstroidBuilder

from python_ta.typecheck.base import (
    TypeFailFunction,
    _collect_tvars,
    _gorg,
    _node_to_type,
    accept_failable,
    class_callable,
    parse_annotations,
)

TYPE_SHED_PATH = os.path.join(os.path.dirname(__file__), "typeshed", "builtins.pyi")


class TypeStore:
    """A representation of the types the entities defined in the current environment."""

    def __init__(self, type_constraints) -> None:
        """Initialize a type store with all the built-in types from the typeshed module."""
        self.type_constraints = type_constraints
        self.classes = defaultdict(lambda: defaultdict(list))
        self.functions = defaultdict(list)
        self.methods = defaultdict(list)

        builder = AstroidBuilder()
        module = builder.file_build(TYPE_SHED_PATH)

        self._parse_classes(module)
        self._parse_functions(module)

        # Add in initializers
        for klass_name, methods in self.classes.items():
            if "__init__" in methods:
                self.functions[klass_name] = [
                    class_callable(init) for init, _ in methods["__init__"]
                ]

    def _parse_classes(self, module: nodes.Module) -> None:
        """Parse the class definitions from typeshed."""
        for class_def in module.nodes_of_class(nodes.ClassDef):
            tvars = []
            self.classes[class_def.name]["__bases"] = []
            for base in class_def.bases:
                base_type = _node_to_type(base)
                self.classes[class_def.name]["__pyta_tvars"] = [
                    tv.__name__ for tv in _collect_tvars(base_type)
                ]
                self.classes[class_def.name]["__bases"].append(base_type)
            self.classes[class_def.name]["__mro"] = [cls.name for cls in class_def.mro()]
            for node in (
                nodes[0]
                for nodes in class_def.locals.values()
                if isinstance(nodes[0], nodes.AssignName)
                and isinstance(nodes[0].parent, nodes.AnnAssign)
            ):
                self.classes[class_def.name][node.name] = parse_annotations(node, tvars)

    def _parse_functions(self, module: nodes.Module) -> None:
        """Parse the function definitions from typeshed."""
        for function_def in module.nodes_of_class(nodes.FunctionDef):
            in_class = isinstance(function_def.parent, nodes.ClassDef)
            if in_class:
                tvars = self.classes[function_def.parent.name]["__pyta_tvars"]
            else:
                tvars = []
            f_type = parse_annotations(function_def, tvars)
            if in_class:
                self.classes[function_def.parent.name][function_def.name].extend(f_type)
                self.methods[function_def.name].extend(f_type)
            else:
                self.functions[function_def.name].extend(f_type)

    def lookup_function(self, operator, *args):
        """Helper method to lookup a function type given the operator and types of arguments."""
        if args:
            func_types_list = self.functions[operator]
            for func_type in func_types_list:
                if len(args) != len(func_type.__args__) - 1:
                    continue
                if self.type_constraints.can_unify(
                    Callable[list(func_type.__args__[:-1]), Any], Callable[list(args), Any]
                ):
                    return func_type
            raise KeyError

    @accept_failable
    def lookup_method(self, operator, *args, node):
        """Helper method to lookup a method type given the operator and types of arguments.

        TODO: modify this implementation to use mro.
        """
        if args:
            # First try to do a direct lookup.
            if (
                isinstance(args[0], ForwardRef)
                and operator in self.classes[args[0].__forward_arg__]
            ):
                for func_type, _ in self.classes[args[0].__forward_arg__][operator]:
                    if len(args) != len(func_type.__args__) - 1:
                        continue
                    if self.type_constraints.can_unify(
                        Callable[list(args), Any], Callable[list(func_type.__args__[:-1]), Any]
                    ):
                        return func_type

            # If that doesn't work, fall back on a brute force search.
            func_types_list = self.methods[operator]
            for func_type, _ in func_types_list:
                if len(args) != len(func_type.__args__) - 1:
                    continue
                if self.type_constraints.can_unify(
                    Callable[list(args), Any], Callable[list(func_type.__args__[:-1]), Any]
                ):
                    return func_type
            return TypeFailFunction(tuple(func_types_list), None, node)

    def is_descendant(self, child: type, ancestor: type) -> bool:
        if (
            getattr(child, "__origin__", None) is Union
            or getattr(ancestor, "__origin__", None) is Union
        ):
            return self.type_constraints.can_unify(child, ancestor)
        if ancestor == object or ancestor == Any or child == Any:
            return True

        if isinstance(child, ForwardRef):
            child_name = child.__forward_arg__
        elif hasattr(child, "__origin__"):
            child_name = child.__origin__.__name__
        else:
            child_name = child.__name__

        if hasattr(child, "__mro__") and ancestor in child.__mro__:
            return True
        elif (
            hasattr(ancestor, "__mro__")
            and Protocol in ancestor.__mro__
            and issubclass(child, ancestor)
        ):
            return True
        elif hasattr(child, "__orig_bases__"):
            for base in child.__orig_bases__:
                if (
                    isinstance(base, _GenericAlias)
                    and isinstance(ancestor, _GenericAlias)
                    and issubclass(_gorg(base), _gorg(ancestor))
                    and all(
                        [
                            self.type_constraints.can_unify(a1, a2)
                            for a1, a2 in zip(base.__args__, ancestor.__args__)
                        ]
                    )
                ):
                    return True
        elif child_name in self.classes:
            for base in self.classes[child_name]["__bases"]:
                if self.type_constraints.can_unify(base, ancestor) or self.is_descendant(
                    base, ancestor
                ):
                    # TODO: make sure it is safe to alter type_constraints here
                    self.type_constraints.unify(base, ancestor)
                    return True
        return False


if __name__ == "__main__":
    # Display the TypeStore parsed from typeshed.
    ts = TypeStore(None)
    import pprint

    pprint.pprint(ts.classes)
    pprint.pprint(ts.methods["__iter__"])
    pprint.pprint(ts.functions)



"""Print out the source text corresponding to AST nodes.
"""
import os

import astroid
import colorama
import inflection
from colorama import Back, Fore, Style

import python_ta.transforms.setendings as setendings

colorama.init(strip=False, autoreset=True)


def _wrap_color(code_string):
    """Wrap key parts in styling and resets.
    Stying for each key part from,
    (col_offset, fromlineno) to (end_col_offset, end_lineno).
    Note: use this to set color back to default (on mac, and others?):
          Style.RESET_ALL + Style.DIM
    """
    ret = Style.BRIGHT + Fore.WHITE + Back.BLACK
    ret += code_string
    ret += Style.RESET_ALL + Style.DIM + Fore.RESET + Back.RESET
    return ret


def print_node(filename, node_class):
    """Print all nodes of the given class in the given file."""
    with open(filename) as f:
        content = f.read()
    source_lines = content.split("\n")
    module = astroid.parse(content)

    # Set end_lineno and end_col_offset for all nodes in `module`.
    ending_transformer = setendings.init_register_ending_setters(source_lines)
    ending_transformer.visit(module)

    for node in module.nodes_of_class(node_class):
        if node.fromlineno == node.end_lineno:
            line = source_lines[node.fromlineno - 1]  # string
            out = [
                line[: node.col_offset]
                +
                # The key part:
                _wrap_color(line[node.col_offset : node.end_col_offset])
                + line[node.end_col_offset :]
            ]
        else:
            first_line = source_lines[node.fromlineno - 1]  # string
            middle_lines = source_lines[node.fromlineno : node.end_lineno - 1]  # list
            last_line = source_lines[node.end_lineno - 1]  # string

            if middle_lines:
                # For each item in the list of lines of strings,
                # add colorama style to middle like the first and last lines
                middle_lines = "\n".join(list(map(_wrap_color, middle_lines))) + "\n"
            else:
                middle_lines = ""  # coerce list to string

            if first_line:  # Add a spacing after first_line
                middle_lines = "\n" + middle_lines

            out = [
                first_line[: node.col_offset]
                +
                # The key part:
                _wrap_color(first_line[node.col_offset :])
                + middle_lines
                + _wrap_color(last_line[: node.end_col_offset])
                + last_line[node.end_col_offset :]
            ]
        print(Style.DIM + "\n".join(out))


if __name__ == "__main__":
    for node_class in astroid.nodes.ALL_NODE_CLASSES:
        print("=== {} ===".format(node_class.__name__))
        file_location = "nodes/" + inflection.underscore(node_class.__name__) + ".py"
        try:
            print_node(file_location, node_class)
        except FileNotFoundError:
            print("WARNING: No file for class {}".format(node_class))
        except AttributeError:
            print("ERROR: for class {}".format(node_class))

        print("")



from typing import Dict, List, Optional, Tuple, Union

from astroid import nodes

from .graph import CFGBlock, ControlFlowGraph


class CFGVisitor:
    """An astroid visitor that creates a control flow graph for a given Python module.

    Private Attributes:
    _control_boundaries: A stack of the boundaries the visitor is currently in.
        The top of the stack corresponds to the end of the list.
        (compound statement [while], {'Break'/'Continue': CFGBlock to link to})
    """

    cfgs: Dict[Union[nodes.FunctionDef, nodes.Module], ControlFlowGraph]
    _current_cfg: Optional[ControlFlowGraph]
    _current_block: Optional[CFGBlock]
    _control_boundaries: List[Tuple[nodes.NodeNG, Dict[str, CFGBlock]]]

    def __init__(self) -> None:
        super().__init__()
        self.cfgs = {}
        self._current_cfg = None
        self._current_block = None
        self._control_boundaries = []

    def __getattr__(self, attr: str):
        if attr.startswith("visit_"):
            return self.visit_generic
        else:
            raise AttributeError(f"'CFGVisitor' object has not attribute '{attr}'")

    def visit_generic(self, node: nodes.NodeNG) -> None:
        """By default, add the expression to the end of the current block."""
        self._current_block.add_statement(node)

    def visit_module(self, module: nodes.Module) -> None:
        self.cfgs[module] = ControlFlowGraph()
        self._current_cfg = self.cfgs[module]
        self._current_block = self._current_cfg.start
        module.cfg_block = self._current_cfg.start

        for child in module.body:
            child.accept(self)

        self._current_cfg.link_or_merge(self._current_block, self._current_cfg.end)
        self._current_cfg.update_block_reachability()

    def visit_classdef(self, node: nodes.ClassDef) -> None:
        for child in node.body:
            child.accept(self)

    def visit_functiondef(self, func: nodes.FunctionDef) -> None:
        self._current_block.add_statement(func)

        previous_cfg = self._current_cfg
        previous_block = self._current_block

        self.cfgs[func] = ControlFlowGraph()
        self._current_cfg = self.cfgs[func]

        self._control_boundaries.append((func, {nodes.Return.__name__: self._current_cfg.end}))

        self._current_cfg.start.add_statement(func.args)
        func.cfg_block = self._current_cfg.start

        self._current_block = self._current_cfg.create_block(self._current_cfg.start)

        for child in func.body:
            child.accept(self)

        self._control_boundaries.pop()

        self._current_cfg.link_or_merge(self._current_block, self._current_cfg.end)
        self._current_cfg.update_block_reachability()

        self._current_block = previous_block
        self._current_cfg = previous_cfg

    def visit_if(self, node: nodes.If) -> None:
        self._current_block.add_statement(node.test)
        node.cfg_block = self._current_block
        old_curr = self._current_block

        # Handle "then" branch.
        then_block = self._current_cfg.create_block(old_curr)
        self._current_block = then_block
        for child in node.body:
            child.accept(self)
        end_if = self._current_block

        # Handle "else" branch.
        if node.orelse == []:
            end_else = old_curr
        else:
            else_block = self._current_cfg.create_block(old_curr)
            self._current_block = else_block
            for child in node.orelse:
                child.accept(self)
            end_else = self._current_block

        after_if_block = self._current_cfg.create_block()
        self._current_cfg.link_or_merge(end_if, after_if_block)
        self._current_cfg.link_or_merge(end_else, after_if_block)

        self._current_block = after_if_block

    def visit_while(self, node: nodes.While) -> None:
        old_curr = self._current_block

        # Handle "test" block
        test_block = self._current_cfg.create_block()
        test_block.add_statement(node.test)
        node.cfg_block = test_block
        self._current_cfg.link_or_merge(old_curr, test_block)

        after_while_block = self._current_cfg.create_block()

        # step into while
        self._control_boundaries.append(
            (
                node,
                {nodes.Break.__name__: after_while_block, nodes.Continue.__name__: test_block},
            )
        )

        # Handle "body" branch
        body_block = self._current_cfg.create_block(test_block)
        self._current_block = body_block
        for child in node.body:
            child.accept(self)
        end_body = self._current_block
        self._current_cfg.link_or_merge(end_body, test_block)

        # step out of while
        self._control_boundaries.pop()

        # Handle "else" branch
        else_block = self._current_cfg.create_block(test_block)
        self._current_block = else_block
        for child in node.orelse:
            child.accept(self)
        end_else = self._current_block

        self._current_cfg.link_or_merge(end_else, after_while_block)
        self._current_block = after_while_block

    def visit_for(self, node: nodes.For) -> None:
        old_curr = self._current_block
        old_curr.add_statement(node.iter)
        node.cfg_block = old_curr

        # Handle "test" block
        test_block = self._current_cfg.create_block()
        test_block.add_statement(node.target)
        self._current_cfg.link_or_merge(old_curr, test_block)

        after_for_block = self._current_cfg.create_block()

        # step into for
        self._control_boundaries.append(
            (node, {nodes.Break.__name__: after_for_block, nodes.Continue.__name__: test_block})
        )

        # Handle "body" branch
        body_block = self._current_cfg.create_block(test_block)
        self._current_block = body_block
        for child in node.body:
            child.accept(self)
        end_body = self._current_block
        self._current_cfg.link_or_merge(end_body, test_block)

        # step out of for
        self._control_boundaries.pop()

        # Handle "else" branch
        else_block = self._current_cfg.create_block(test_block)
        self._current_block = else_block
        for child in node.orelse:
            child.accept(self)
        end_else = self._current_block

        self._current_cfg.link_or_merge(end_else, after_for_block)
        self._current_block = after_for_block

    def visit_break(self, node: nodes.Break) -> None:
        self._visit_jump(node)

    def visit_continue(self, node: nodes.Continue) -> None:
        self._visit_jump(node)

    def visit_return(self, node: nodes.Return) -> None:
        self._visit_jump(node)

    def _visit_jump(self, node: Union[nodes.Break, nodes.Continue, nodes.Return]) -> None:
        old_curr = self._current_block
        for boundary, exits in reversed(self._control_boundaries):
            if type(node).__name__ in exits:
                self._current_cfg.link(old_curr, exits[type(node).__name__])
                old_curr.add_statement(node)
                break
        else:
            raise SyntaxError(
                f"'{type(node).__name__}' outside"
                f' {"function" if isinstance(node, nodes.Return) else "loop"}'
            )
        unreachable_block = self._current_cfg.create_block()
        self._current_block = unreachable_block

    def visit_tryexcept(self, node: nodes.TryExcept) -> None:
        if self._current_block.statements != []:
            self._current_block = self._current_cfg.create_block(self._current_block)

        node.cfg_block = self._current_block

        for child in node.body:
            child.accept(self)
        end_body = self._current_block

        end_block = self._current_cfg.create_block()

        after_body = []
        for handler in node.handlers:
            h = self._current_cfg.create_block()
            self._current_block = h
            handler.cfg_block = h
            if handler.name is not None:  # The name assigned to the caught exception.
                handler.name.accept(self)
            for child in handler.body:
                child.accept(self)
            end_handler = self._current_block
            self._current_cfg.link_or_merge(end_handler, end_block)
            after_body.append(h)

        if node.orelse == []:
            after_body.append(end_block)
        else:
            self._current_block = self._current_cfg.create_block()
            after_body.append(self._current_block)
            for child in node.orelse:
                child.accept(self)
            self._current_cfg.link_or_merge(self._current_block, end_block)

        self._current_cfg.multiple_link_or_merge(end_body, after_body)
        self._current_block = end_block

    def visit_with(self, node: nodes.With) -> None:
        for context_node, name in node.items:
            self._current_block.add_statement(context_node)
            if name is not None:
                self._current_block.add_statement(name)

        for child in node.body:
            child.accept(self)



"""checker for type annotation.
"""

from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.checkers.utils import check_messages
from pylint.interfaces import IAstroidChecker


class TypeAnnotationChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = "TypeAnnotationChecker"
    msgs = {
        "E9970": (
            "Function parameter is missing type annotation",
            "missing-param-type",
            "Presented when a type annotation is missing.",
        ),
        "E9971": (
            "Function is missing return type annotation",
            "missing-return-type",
            "Presented when a type annotation is missing.",
        ),
        "E9972": (
            "Attribute is missing type annotation in class body",
            "missing-attribute-type",
            "Presented when a type annotation is missing.",
        ),
    }

    # this is important so that your checker is executed before others
    priority = -1

    @check_messages("missing-param-type", "missing-return-type")
    def visit_functiondef(self, node):
        for i in range(len(node.args.annotations)):
            if node.args.annotations[i] is None:
                # Check if function is a non-static instance method
                if (
                    i != 0
                    or not isinstance(node.parent, nodes.ClassDef)
                    or node.type == "staticmethod"
                ):
                    self.add_message("missing-param-type", node=node.args.args[i])

        if node.returns is None:
            self.add_message("missing-return-type", node=node.args)

    @check_messages("missing-attribute-type")
    def visit_classdef(self, node):
        for attr_key in node.instance_attrs:
            attr_node = node.instance_attrs[attr_key][0]
            if isinstance(attr_node, nodes.AssignAttr):
                if attr_key not in node.locals and all(
                    attr_key not in base.locals for base in node.ancestors()
                ):
                    self.add_message("missing-attribute-type", node=attr_node)
                elif isinstance(attr_node.parent, nodes.AnnAssign):
                    self.add_message("missing-attribute-type", node=attr_node)

        for attr_key in node.locals:
            attr_node = node.locals[attr_key][0]
            if isinstance(attr_node, nodes.AssignName) and not isinstance(
                attr_node.parent, nodes.AnnAssign
            ):
                self.add_message("missing-attribute-type", node=attr_node)


def register(linter):
    linter.register_checker(TypeAnnotationChecker(linter))



import hypothesis.strategies as hs
from astroid import nodes
from hypothesis import HealthCheck, assume, given, settings

from python_ta.typecheck.base import TypeFail

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


@given(
    hs.integers(), hs.lists(hs.tuples(cs.comparator_operator_equality, hs.integers()), min_size=1)
)
@settings(deadline=None)
def test_compare_equality(left_value, operator_value_tuples):
    """Test type setting of Compare node representing comparators: ''==', '!=', '>=', '<=', 'is'."""
    program = f"{repr(left_value)}"
    for operator, value in operator_value_tuples:
        program += " " + " ".join([operator, repr(value)])
    module, _ = cs._parse_text(program)
    compare_node = list(module.nodes_of_class(nodes.Compare))[0]
    assert compare_node.inf_type.getValue() == bool


@given(hs.lists(cs.comparator_operator, min_size=3), cs.numeric_list(min_size=4))
@settings(deadline=None)
def test_compare_inequality(operators, values):
    """Test type setting of Compare node representing comparators: '<', '>'."""
    a = list(zip(operators, values))
    pre = []
    for operator, value in a:
        pre.append(str(operator))
        pre.append(str(value))
    # pre_input_program = [str(elt) for tuple in zip(operator, values) for elt in tuple]
    program = f"{str(values[0])} " + " ".join(pre)
    module, _ = cs._parse_text(program)
    compare_node = list(module.nodes_of_class(nodes.Compare))[0]
    assert compare_node.inf_type.getValue() == bool


def test_compare_in():
    """Test type setting of Compare node representing 'in' operator"""
    program = "0 in [1,2,3]"
    module, _ = cs._parse_text(program)
    compare_node = list(module.nodes_of_class(nodes.Compare))[0]
    assert compare_node.inf_type.getValue() == bool


def test_compare_in_error():
    """Test type setting of Compare node representing invaliud use of 'in' operator"""
    program = "0 in 5"
    module, _ = cs._parse_text(program)
    compare_node = list(module.nodes_of_class(nodes.Compare))[0]
    assert isinstance(compare_node.inf_type, TypeFail)


def test_compare_is():
    program = """
    A = 0
    A is 1
    """
    module, _ = cs._parse_text(program)
    compare_node = list(module.nodes_of_class(nodes.Compare))[0]
    assert compare_node.inf_type.getValue() == bool


def test_compare_is_not():
    program = """
    A = 0
    A is not 1
    """
    module, _ = cs._parse_text(program)
    compare_node = list(module.nodes_of_class(nodes.Compare))[0]
    assert compare_node.inf_type.getValue() == bool


def test_compare_not_in():
    program = """
    A = [0, 1, 2]
    3 not in A

    """
    module, _ = cs._parse_text(program)
    compare_node = list(module.nodes_of_class(nodes.Compare))[0]
    assert compare_node.inf_type.getValue() == bool


def test_compare_not_in_fail():
    program = """
    A = 5
    3 not in A

    """
    module, _ = cs._parse_text(program)
    compare_node = list(module.nodes_of_class(nodes.Compare))[0]
    assert isinstance(compare_node.inf_type, TypeFail)



from keyword import iskeyword
from typing import Callable

import hypothesis.strategies as hs
from astroid import nodes
from hypothesis import HealthCheck, assume, given, settings
from pytest import skip

from python_ta.transforms.type_inference_visitor import (
    TypeFail,
    TypeFailFunction,
    TypeFailLookup,
)

from .. import custom_hypothesis_support as cs
from ..custom_hypothesis_support import types_in_callable

settings.load_profile("pyta")


def _parse_to_function(function_name, args_list, return_statement):
    """Helper to parse given data into function definition."""
    return f'def {function_name}({", ".join(args_list)}):' f"    return {return_statement}"


def _parse_to_function_no_return(function_name, args_list, function_body):
    """Helper to parse given data into function definition."""
    return f'def {function_name}({", ".join(args_list)}):\n' f"    {function_body}"


@given(cs.valid_identifier(), cs.primitive_values)
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_function_def_no_args(function_name, return_value):
    """Test FunctionDef node visitors representing function definitions with no parameters and primitive return type."""
    assume(not iskeyword(function_name))
    program = _parse_to_function(function_name, [], repr(return_value))
    module, inferer = cs._parse_text(program)
    function_type_var = module.type_environment.lookup_in_env(function_name)
    function_def_node = list(module.nodes_of_class(nodes.FunctionDef))[0]
    return_tvar = function_def_node.type_environment.lookup_in_env("return")
    return_type = inferer.type_constraints.resolve(return_tvar).getValue()
    assert (
        inferer.type_constraints.resolve(function_type_var).getValue() == Callable[[], return_type]
    )


@given(cs.valid_identifier(), cs.primitive_values)
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_function_def_call_no_args(function_name, return_value):
    """Test type setting in environment of a function call for a function with no parameters."""
    program = (
        _parse_to_function(function_name, [], repr(return_value)) + "\n" + function_name + "()\n"
    )
    module, inferer = cs._parse_text(program)
    # there should be a single Expr node in this program
    function_def_node = list(module.nodes_of_class(nodes.FunctionDef))[0]
    return_tvar = function_def_node.type_environment.lookup_in_env("return")
    return_type = inferer.type_constraints.resolve(return_tvar).getValue()
    expr_node = next(module.nodes_of_class(nodes.Expr))
    assert expr_node.inf_type.getValue() == return_type


@given(cs.valid_identifier(), hs.lists(cs.valid_identifier(), min_size=0), cs.primitive_values)
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_function_def_no_return(function_name, arguments, body):
    """Test FunctionDef node visitors representing non-returning function definitions with parameter(s)."""
    for return_value in ["return None", repr(body), "pass"]:
        program = _parse_to_function_no_return(function_name, arguments, repr(return_value))
        module, inferer = cs._parse_text(program)
        function_def_node = next(module.nodes_of_class(nodes.FunctionDef))
        expected_arg_type_vars = [
            function_def_node.type_environment.lookup_in_env(argument) for argument in arguments
        ]
        expected_arg_types = [
            inferer.type_constraints.resolve(type_var).getValue()
            for type_var in expected_arg_type_vars
        ]
        function_type_var = module.type_environment.lookup_in_env(function_name)
        assert (
            inferer.type_constraints.resolve(function_type_var).getValue()
            == Callable[expected_arg_types, None]
        )


@given(cs.valid_identifier(), hs.lists(cs.valid_identifier(), min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow], deadline=None)
def test_function_def_args_simple_return(function_name, arguments):
    """Test FunctionDef node visitors representing function definitions with paramater(s):
    return one of its arguments."""
    for argument in arguments:
        program = _parse_to_function(function_name, arguments, argument)
        module, inferer = cs._parse_text(program)
        # get the functionDef node - there is only one in this test case.
        function_def_node = next(module.nodes_of_class(nodes.FunctionDef))
        expected_arg_type_vars = [
            function_def_node.type_environment.lookup_in_env(argument) for argument in arguments
        ]
        expected_arg_types = [
            inferer.type_constraints.resolve(type_var).getValue()
            for type_var in expected_arg_type_vars
        ]
        function_type_var = module.type_environment.lookup_in_env(function_name)
        function_type = inferer.type_constraints.resolve(function_type_var).getValue()
        actual_arg_types, actual_return_type = types_in_callable(inferer, function_type)
        return_type_var = function_def_node.type_environment.lookup_in_env(argument)
        expected_return_type = inferer.type_constraints.resolve(return_type_var).getValue()
        assert (
            Callable[actual_arg_types, actual_return_type]
            == Callable[expected_arg_types, expected_return_type]
        )


@given(cs.valid_identifier(), cs.random_dict_variable_homogeneous_value(min_size=1, max_size=4))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_function_def_args_simple_function_call(function_name, variables_dict):
    """Test setting f env for function call of function definitions with params that returns one of it's arguments."""
    assume(not iskeyword(function_name) and function_name not in variables_dict)
    # generate every possible function definition program of aforementioned form.
    for i in range(len(list(variables_dict.keys()))):
        return_value = list(variables_dict.keys())[i]
        arguments = ", ".join([repr(value) for value in variables_dict.values()])
        program = (
            f"{_parse_to_function(function_name, list(variables_dict.keys()), return_value)}\n"
            f"{function_name}({arguments})"
        )
        module, inferer = cs._parse_text(program)
        # get the Call node - there is only one in this test case.
        call_node = next(module.nodes_of_class(nodes.Call))
        function_call_type = call_node.inf_type.getValue()
        assert (
            inferer.type_constraints.resolve(function_call_type).getValue()
            == call_node.args[i].inf_type.getValue()
        )


from keyword import iskeyword
from typing import Callable

import astroid
import hypothesis.strategies as hs
from hypothesis import HealthCheck, assume, given, settings
from pytest import skip

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


def _parse_to_function(function_name, args_list, return_statement):
    """Helper to parse given data into function definition."""
    return f'def {function_name}({", ".join(args_list)}):' f"    return {return_statement}"


def _parse_to_function_no_return(function_name, args_list, function_body):
    """Helper to parse given data into function definition."""
    return f'def {function_name}({", ".join(args_list)}):\n' f"    {function_body}"


@given(cs.valid_identifier(), cs.primitive_values)
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_function_def_no_args(function_name, return_value):
    """Test FunctionDef node visitors representing function definitions with no parameters and primitive return type."""
    assume(not iskeyword(function_name))
    program = _parse_to_function(function_name, [], repr(return_value))
    module, inferer = cs._parse_text(program)
    function_type_var = module.type_environment.lookup_in_env(function_name)
    function_def_node = list(module.nodes_of_class(nodes.FunctionDef))[0]
    return_tvar = function_def_node.type_environment.lookup_in_env("return")
    return_type = inferer.type_constraints.resolve(return_tvar).getValue()
    assert (
        inferer.type_constraints.resolve(function_type_var).getValue() == Callable[[], return_type]
    )


@given(cs.valid_identifier(), cs.primitive_values)
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_function_def_call_no_args(function_name, return_value):
    """Test type setting in environment of a function call for a function with no parameters."""
    program = (
        _parse_to_function(function_name, [], repr(return_value)) + "\n" + function_name + "()\n"
    )
    module, inferer = cs._parse_text(program)
    # there should be a single Expr node in this program
    function_def_node = list(module.nodes_of_class(nodes.FunctionDef))[0]
    return_tvar = function_def_node.type_environment.lookup_in_env("return")
    return_type = inferer.type_constraints.resolve(return_tvar).getValue()
    expr_node = next(module.nodes_of_class(nodes.Expr))
    assert expr_node.inf_type.getValue() == return_type


@given(cs.valid_identifier(), hs.lists(cs.valid_identifier(), min_size=0), cs.primitive_values)
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_function_def_no_return(function_name, arguments, body):
    """Test FunctionDef node visitors representing non-returning function definitions with parameter(s)."""
    for return_value in ["return None", repr(body), "pass"]:
        program = _parse_to_function_no_return(function_name, arguments, repr(return_value))
        module, inferer = cs._parse_text(program)
        function_def_node = next(module.nodes_of_class(nodes.FunctionDef))
        expected_arg_type_vars = [
            function_def_node.type_environment.lookup_in_env(argument) for argument in arguments
        ]
        expected_arg_types = [
            inferer.type_constraints.resolve(type_var).getValue()
            for type_var in expected_arg_type_vars
        ]
        function_type_var = module.type_environment.lookup_in_env(function_name)
        assert (
            inferer.type_constraints.resolve(function_type_var).getValue()
            == Callable[expected_arg_types, None]
        )


@given(cs.valid_identifier(), hs.lists(cs.valid_identifier(), min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_function_def_args_simple_return(function_name, arguments):
    """Test FunctionDef node visitors representing function definitions with paramater(s):
    return one of its arguments."""
    for argument in arguments:
        program = _parse_to_function(function_name, arguments, argument)
        module, inferer = cs._parse_text(program)
        # get the functionDef node - there is only one in this test case.
        function_def_node = next(module.nodes_of_class(nodes.FunctionDef))
        expected_arg_type_vars = [
            function_def_node.type_environment.lookup_in_env(argument) for argument in arguments
        ]
        expected_arg_types = [
            inferer.type_constraints.resolve(type_var).getValue()
            for type_var in expected_arg_type_vars
        ]
        function_type_var = module.type_environment.lookup_in_env(function_name)
        function_type = inferer.type_constraints.resolve(function_type_var).getValue()
        actual_arg_types, actual_return_type = types_in_callable(inferer, function_type)
        return_type_var = function_def_node.type_environment.lookup_in_env(argument)
        expected_return_type = inferer.type_constraints.resolve(return_type_var).getValue()
        assert inferer.type_constraints.can_unify(
            Callable[actual_arg_types, actual_return_type],
            Callable[expected_arg_types, expected_return_type],
        )


@given(cs.valid_identifier(), cs.random_dict_variable_homogeneous_value(min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_function_def_args_simple_function_call(function_name, variables_dict):
    """Test setting f env for function call of function definitions with params that returns one of it's arguments."""
    assume(not iskeyword(function_name) and function_name not in variables_dict)
    # generate every possible function definition program of aforementioned form.
    for i in range(len(list(variables_dict.keys()))):
        return_value = list(variables_dict.keys())[i]
        arguments = ", ".join([repr(value) for value in variables_dict.values()])
        program = (
            f"{_parse_to_function(function_name, list(variables_dict.keys()), return_value)}\n"
            f"{function_name}({arguments})"
        )
        module, inferer = cs._parse_text(program)
        # get the Call node - there is only one in this test case.
        call_node = next(module.nodes_of_class(nodes.Call))
        function_call_type = call_node.inf_type.getValue()
        assert (
            inferer.type_constraints.resolve(function_call_type).getValue()
            == call_node.args[i].inf_type.getValue()
        )


def test_non_annotated_function_call_bad_arguments():
    """User tries to call a non-annotated function on arguments of the wrong type."""
    program = (
        f"def add_num(num1, num2):\n" f"    return num1 + num2\n" f"\n" f'add_num("bob", 1.0)\n'
    )
    try:
        module, inferer = cs._parse_text(program)
    except:
        skip()
    call_node = next(module.nodes_of_class(nodes.Call))
    # TODO: This error is flawed because the unification error occurs for both arguments due to our current implementation,
    # which "chooses" the first valid function type from TypeStore.
    # Should we fix this implementation first or save it for later and hard-code the correct error message for now?
    expected_msg = (
        f'In the Call node in line 4, there was an error in calling the function "add_num":\n'
        f"in parameter (1), the function was expecting an object of inferred type "
        f"int but was given an object of type str.\n"
        f"in parameter (2), the function was expecting an object of inferred type "
        f"int but was given an object of type float.\n"
    )
    # TODO: should we use the term inferred?
    assert call_node.inf_type.getValue() == expected_msg


def test_user_defined_annotated_call_wrong_arguments_type():
    """User tries to call an annotated user-defined function on the wrongly-typed arguments."""
    program = (
        f"def add_3(num1: int, num2: int, num3: int) -> int:\n"
        f"    return num1 + num2 + num3\n"
        f"\n"
        f'add_3(1, "bob", 1.0)\n'
    )
    try:
        module, inferer = cs._parse_text(program)
    except:
        skip()
    call_node = list(module.nodes_of_class(nodes.Call))[0]
    expected_msg = (
        f'In the Call node in line 4, there was an error in calling the annotated function "add_3":\n'
        f"in parameter (2), the annotated type is int but was given an object of type str.\n"
        f"in parameter (3), the annotated type is int but was given an object of type float.\n"
    )
    assert call_node.inf_type.getValue() == expected_msg


def test_user_defined_annotated_call_wrong_arguments_number():
    """User tries to call an annotated function on the wrong number of arguments."""
    program = (
        f"def add_3(num1: int, num2: int, num3: int) -> int:\n"
        f"    return num1 + num2 + num3\n"
        f"\n"
        f"add_3()\n"
    )
    try:
        module, inferer = cs._parse_text(program)
    except:
        skip()
    call_node = list(module.nodes_of_class(nodes.Call))[0]
    expected_msg = (
        f'In the Call node in line 4, there was an error in calling the function "add_3":\n'
        f"the function was expecting 3 arguments, but was given 0."
    )
    assert call_node.inf_type.getValue() == expected_msg


def test_conflicting_inferred_type_variable():
    """User calls two functions on an object, which contradicts the inferred type of the variable."""
    program = """
        def return_num(num: int) -> int:
            return num

        def return_str(str: str) -> str:
            return str

        def f(x):
            return_num(x)
            return_str(x)
        """
    try:
        module, inferer = cs._parse_text(program)
    except:
        skip()
    call_node = list(module.nodes_of_class(nodes.Call))[1]
    expected_msg = (
        f'In the Call node in line 8, there was an error in calling the annotated function "return_str":\n'
        f"in parameter (1), the annotated type is str but was given an object of inferred type int."
    )
    # TODO: test case redundant because recursive..?
    assert call_node.inf_type.getValue() == expected_msg


def test_non_annotated_function_call_bad_arguments():
    """User tries to call a non-annotated function on arguments of the wrong type."""
    skip("Skipping this test until error messages are fixed")
    program = (
        f"def add_num(num1, num2):\n" f"    return num1 + num2\n" f"\n" f'add_num("bob", 1.0)\n'
    )
    try:
        module, inferer = cs._parse_text(program)
    except:
        skip()
    call_node = next(module.nodes_of_class(nodes.Call))
    # TODO: This error is flawed because the unification error occurs for both arguments due to our current implementation,
    # which "chooses" the first valid function type from TypeStore.
    # Should we fix this implementation first or save it for later and hard-code the correct error message for now?
    expected_msg = (
        f'In the Call node in line 4, there was an error in calling the function "add_num":\n'
        f"in parameter (1), the function was expecting an object of inferred type "
        f"int but was given an object of type str.\n"
        f"in parameter (2), the function was expecting an object of inferred type "
        f"int but was given an object of type float.\n"
    )
    # TODO: should we use the term inferred?
    assert call_node.inf_type.getValue() == expected_msg


def test_user_defined_annotated_call_wrong_arguments_type():
    """User tries to call an annotated user-defined function on the wrongly-typed arguments."""
    skip("Skipping this test until error messages are fixed")
    program = (
        f"def add_3(num1: int, num2: int, num3: int) -> int:\n"
        f"    return num1 + num2 + num3\n"
        f"\n"
        f'add_3(1, "bob", 1.0)\n'
    )
    try:
        module, inferer = cs._parse_text(program)
    except:
        skip()
    call_node = list(module.nodes_of_class(nodes.Call))[0]
    expected_msg = (
        f'In the Call node in line 4, there was an error in calling the annotated function "add_3":\n'
        f"in parameter (2), the annotated type is int but was given an object of type str.\n"
        f"in parameter (3), the annotated type is int but was given an object of type float.\n"
    )
    assert call_node.inf_type.getValue() == expected_msg


def test_user_defined_annotated_call_wrong_arguments_number():
    """User tries to call an annotated function on the wrong number of arguments."""
    program = (
        f"def add_3(num1: int, num2: int, num3: int) -> int:\n"
        f"    return num1 + num2 + num3\n"
        f"\n"
        f"add_3()\n"
    )
    try:
        module, inferer = cs._parse_text(program)
    except:
        skip()
    call_node = list(module.nodes_of_class(nodes.Call))[0]
    expected_msg = "Wrong number of arguments"
    assert isinstance(call_node.inf_type, TypeFail)


def test_conflicting_inferred_type_variable():
    """User calls two functions on an object, which contradicts the inferred type of the variable."""
    skip("Skipping this test until error messages are fixed")
    program = """
        def return_num(num: int) -> int:
            return num

        def return_str(str: str) -> str:
            return str

        def f(x):
            return_num(x)
            return_str(x)
        """
    try:
        module, inferer = cs._parse_text(program)
    except:
        skip()
    call_node = list(module.nodes_of_class(nodes.Call))[1]
    expected_msg = (
        f'In the Call node in line 8, there was an error in calling the annotated function "return_str":\n'
        f"in parameter (1), the annotated type is str but was given an object of inferred type int."
    )
    # TODO: test case redundant because recursive..?
    assert call_node.inf_type.getValue() == expected_msg


def test_non_callable():
    program = """
    x = 1
    x()
    """
    module, inferer = cs._parse_text(program)
    call_node = next(module.nodes_of_class(nodes.Call))
    assert isinstance(call_node.inf_type, TypeFailFunction)


def test_magic_call():
    program = """
    class A:
        def __init__(self):
            self.attr = 0

        def __call__(self):
            return self.attr

    a = A()
    a()
    a.__call__()
    """
    module, inferer = cs._parse_text(program, reset=True)
    for call_node in list(module.nodes_of_class(nodes.Call))[1:]:
        assert call_node.inf_type.getValue() == int


def test_no_magic_call():
    program = """
    class A:
        def __init__(self):
            self.attr = 0

    a = A()
    a()
    a.__call__()
    """
    module, inferer = cs._parse_text(program, reset=True)
    for call_node in list(module.nodes_of_class(nodes.Call))[1:]:
        assert isinstance(call_node.inf_type, TypeFailLookup)


def test_magic_call_wrong_args():
    program = """
    class A:
        def __init__(self):
            self.attr = 0

        def __call__(self, x: int):
            return self.attr + x

    a = A()
    a()
    a.__call__()
    """
    module, inferer = cs._parse_text(program, reset=True)
    for call_node in list(module.nodes_of_class(nodes.Call))[1:]:
        assert isinstance(call_node.inf_type, TypeFailFunction)



import sys
from typing import Dict, List, Set

import pytest

from python_ta.contracts import check_contracts


def test_nullary_return_int() -> None:
    """Calling a nullary function with the correct return type (int)."""

    @check_contracts
    def nullary() -> int:
        return 1

    nullary()


def test_nullary_return_dict() -> None:
    """Calling a nullary function with the correct return type (Dict)."""

    @check_contracts
    def nullary() -> Dict[str, int]:
        return {"one": 1}

    nullary()


def test_nullary_return_none() -> None:
    """Calling a nullary function with the correct return type (None)."""

    @check_contracts
    def nullary() -> None:
        3 + 4

    nullary()


def test_nullary_return_wrong_type() -> None:
    """Calling a nullary function with the incorrect return type."""

    @check_contracts
    def nullary() -> str:
        return 1

    with pytest.raises(AssertionError):
        nullary()


def test_nullary_return_dict_wrong() -> None:
    """Calling a nullary function with the incorrect return type (Dict)."""

    @check_contracts
    def nullary() -> Dict[str, int]:
        return {1: 1}

    with pytest.raises(AssertionError):
        nullary()


def test_nullary_no_return_type() -> None:
    """Calling a nullary function with no specified return type passes."""

    @check_contracts
    def nullary():
        return "string"

    nullary()


@check_contracts
def _my_sum(numbers: List[int]) -> int:
    return sum(numbers)


def test_my_sum_list_int_argument() -> None:
    """Calling _my_sum with a list of integers passes type check."""
    _my_sum([1, 2, 3])


def test_my_sum_empty_list_argument() -> None:
    """Calling _my_sum with an empty list passes type check."""
    _my_sum([])


def test_my_sum_list_mixed_argument() -> None:
    """Calling _my_sum with a list containing not just ints fails type check."""
    with pytest.raises(AssertionError):
        _my_sum([1, 2, "hello"])


@check_contracts
def _my_sum_one_precondition(numbers: List[int]) -> int:
    """Return the sum of a list of numbers.

    Precondition: len(numbers) > 2
    """
    return sum(numbers)


def test_my_sum_one_pre_valid() -> None:
    """Calling _my_sum_one_precondition on a valid input."""
    assert _my_sum_one_precondition([1, 2, 3]) == 6


def test_my_sum_one_wrong_type() -> None:
    """Calling _my_sum_one_precondition on a value of the wrong type."""
    with pytest.raises(AssertionError):
        _my_sum_one_precondition("hi")


def test_my_sum_one_pre_violation() -> None:
    """Calling _my_sum_one_precondition one a value of the right type,
    but that violates the docstring precondition.
    """
    with pytest.raises(AssertionError) as excinfo:
        _my_sum_one_precondition([1])

    msg = str(excinfo.value)
    assert "len(numbers) > 2" in msg


# Checking to see if functions we defined are in-scope for preconditions


def is_even(lst: List[int]) -> bool:
    return all([(not x & 1) for x in lst])


@check_contracts
def _is_even_sum(numbers: List[int]) -> int:
    """Return the sum of a list of numbers.

    Precondition: is_even(numbers)
    """
    return sum(numbers)


def test_is_even_sum_valid() -> None:
    """Calling _is_even_sum on a valid input."""
    assert _is_even_sum([2, 4, 6]) == 12


def test_is_even_sum_violation() -> None:
    """Calling _is_even_sum one a value of the right type,
    but that violates the docstring precondition.
    """
    with pytest.raises(AssertionError) as excinfo:
        _is_even_sum([1, 2])

    msg = str(excinfo.value)
    assert "is_even(numbers)" in msg


@check_contracts
def search(numbers: Set[int]) -> bool:
    """Search for a number in a set.

    Illustrates a preconditions with a double comprehension.

    Preconditions:
        - all({n + m > 0 for n in numbers for m in numbers})
    """
    return 1 in numbers


def test_search_valid() -> None:
    """Test search on a valid input."""
    assert search({1, 2})


def test_search_invalid() -> None:
    """Test search on an invalid input."""
    with pytest.raises(AssertionError) as excinfo:
        search({-1, -2})

    msg = str(excinfo.value)
    assert "all({n + m > 0 for n in numbers for m in numbers})" in msg


class Player:
    user: str


class CPU(Player):
    def __init__(self) -> None:
        self.user = "CPU"


@check_contracts
def _is_cpu(player: Player) -> bool:
    return player.user == "CPU"


def test_class_not_instance_error() -> None:
    """Test that the additional suggestion is added when the class type is passed in as the
    argument instead of its instance

    This test is coupled to the suggestion's arbitrarily chosen text, hence should be updated
    when changing the suggestion text.
    """
    with pytest.raises(AssertionError) as excinfo:
        _is_cpu(Player)

    msg = str(excinfo.value)
    assert "Did you mean Player(...) instead of Player?" in msg


def test_subclass_not_instance_error() -> None:
    """Test that the additional suggestion is added when a subclass type is passed in as an
    argument instead of its instance

    This test is coupled to the suggestion's arbitrarily chosen text, hence should be updated
    when changing the suggestion text.
    """
    with pytest.raises(AssertionError) as excinfo:
        _is_cpu(CPU)

    msg = str(excinfo.value)
    assert "Did you mean CPU(...) instead of CPU?" in msg


def test_no_suggestion_instance_as_instance() -> None:
    """Test that the additional suggestion is not added when an unrelated type is passed in.

    This test is coupled to the suggestion's arbitrarily chosen text, hence should be updated
    when changing the suggestion text.
    """
    with pytest.raises(AssertionError) as excinfo:
        _is_cpu(str)

    msg = str(excinfo.value)

    part1, part2, part3 = "Did you pass in", "instead of", "(...)?"
    assert part1 not in msg
    assert part2 not in msg
    assert part3 not in msg


def test_invalid_typing_generic_argument() -> None:
    """Test that subclass checking on a type parameter that is typing's _GenericAlias does not
    throw an error (as issubclass does not take in a _GenericAlias as its second argument).
    """

    @check_contracts
    def unary(arg: List[str]) -> None:
        return

    with pytest.raises(AssertionError):
        unary(dict)


@pytest.mark.skipif(sys.version_info < (3, 9), reason="built-in generics not yet supported")
def test_invalid_built_in_generic_argument() -> None:
    """Test that subclass checking on a type parameter that is a GenericAlias does not
    throw an error (as issubclass does not take in a GenericAlias as its second argument).
    """

    @check_contracts
    def unary(arg: list[str]) -> None:
        return

    with pytest.raises(AssertionError):
        unary(dict)
